pwf = get_workflow_for('Patch', ask_confirm=False)
deleted = pwf.state_by_name('deleted')
accept = pwf.transition_by_name('accept')
rql('DELETE S allowed_transition T WHERE S eid %(s)s, T eid %(t)s',
    {'s': deleted.eid, 't': accept.eid})

refuse = pwf.transition_by_name('refuse')
refuse.set_attributes(name=u'ask rework')



rql('SET R has_review False WHERE R has_review NULL', ask_confirm=True)
commit()
sync_schema_props_perms('has_review')

sync_schema_props_perms('patch_reviewer')
sync_schema_props_perms('patch_committer')



# copyright 2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-vcreview web ui"""

__docformat__ = "restructuredtext en"
_ = unicode

from cubicweb.web import formwidgets as wdg
from cubicweb.web.views import uicfg

from cubes.vcreview.entities import Patch

_affk = uicfg.autoform_field_kwargs
_abaa = uicfg.actionbox_appearsin_addmenu
_pvs = uicfg.primaryview_section

# insertion point handling #####################################################

_pvs.tag_object_of(('InsertionPoint', 'point_of', '*'), 'hidden')
_abaa.tag_object_of(('InsertionPoint', 'point_of', '*'), False)
_abaa.tag_subject_of(('InsertionPoint', 'has_activity', '*'), False)
_affk.tag_attribute(('InsertionPoint', 'lid'),
                    {'widget': wdg.HiddenInput})
_affk.tag_subject_of(('InsertionPoint', 'point_of', '*'),
                     {'widget': wdg.HiddenInput})


# utility functions ############################################################

def final_patch_states_rql():
    return ' IN (%s)' % ','.join(repr(s) for s in Patch.final_states)



# update workflow condition

add_attribute('Repository', 'has_review')
rql('SET R has_review False', ask_confirm=True)
commit()


rql('Set R repository_committer U '
    'WHERE R has_patch_repository PR, PR repository_committer U,'
    '      NOT R repository_committer U',
    ask_confirm=True)
rql('Set R repository_reviewer U '
    'WHERE R has_patch_repository PR, PR repository_reviewer U,'
    '      NOT R repository_reviewer U',
    ask_confirm=True)

rql('Set R has_review True WHERE R has_patch_repository PR', ask_confirm=True)
commit()
rql('DELETE Repository PR WHERE R has_patch_repository PR', ask_confirm=True)
commit()

drop_relation_type('has_patch_repository')

drop_attribute('Patch', 'branch')
drop_attribute('Patch', 'originator')

drop_relation_type('patch_repository')

add_relation_definition('Patch', 'patch_revision', 'Revision')
drop_relation_definition('Patch', 'patch_revision', 'VersionContent')
drop_relation_definition('Patch', 'patch_revision', 'DeletedVersionContent')

drop_relation_type('applied_at')

drop_relation_type('patch_depends_on')


add_relation_definition('InsertionPoint', 'point_of', 'Revision')
drop_relation_definition('InsertionPoint', 'point_of', 'VersionContent')

add_relation_definition('Revision', 'nosy_list', 'CWUser')
drop_relation_definition('VersionContent', 'nosy_list', 'CWUser')
drop_relation_definition('DeletedVersionContent', 'nosy_list', 'CWUser')

rql('DELETE Patch P', ask_confirm=True)
commit()

# all patches are gone, so no need to keep and update the old wf
wf = get_workflow_for('Patch')
wf.cw_delete()
commit()
from cubes.vcreview.workflows import define_patch_workflow
define_patch_workflow(add_workflow)
commit()

sync_schema_props_perms()



# copyright 2011-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-vcreview startup views"""

__docformat__ = "restructuredtext en"
_ = unicode

import httplib

from logilab.common.decorators import cached
from logilab.mtconverter import xml_escape

from cubicweb import tags, view
from cubicweb.predicates import is_instance, rql_condition, anonymous_user
from cubicweb.web import component, Redirect
from cubicweb.web.views import tableview
from cubicweb.web.views.urlrewrite import SchemaBasedRewriter, rgx, build_rset

from cubes.vcreview.views import final_patch_states_rql

class VCReviewURLRewriter(SchemaBasedRewriter):
    all_patches_title = _('All active patches')
    all_patches_rql = (
           'Any P,P,P,P,PB,PS,R '
           'GROUPBY R,P,PB,PS '
           'ORDERBY RT,PB '
           'WHERE P patch_revision RE, '
           '      RE branch PB, '
           '      P in_state PS,'
           '      RE from_repository R, '
           '      R title RT, '
           '      NOT PS name %s')

    user_worklist_title = _('My review worklist')
    user_worklist_rql = (
           'Any P,P,P,P,PB,PS,R '
           'GROUPBY R,P,PB,PS '
           'ORDERBY RT,PB '
           'WHERE U eid %(u)s, '
           '      P in_state PS, '
           '      P patch_revision RE, '
           '      RE branch PB, '
           '      RE from_repository R, '
           '      R title RT, '
           # Patch is pending review  and CWUser is reviewer
           '      EXISTS(PS name "pending-review" AND P patch_reviewer U) '
           # Patch is reviewed  and CWUser is commiter
           '      OR EXISTS(PS name "reviewed" AND P patch_committer U) '
           # Patch is reviewed  and CWUser repository or global commiter
           '      OR (EXISTS(PS name "reviewed"'
           '          AND NOT EXISTS(P patch_committer V) '
           '          AND (EXISTS(R repository_committer U)'
           '              OR EXISTS(R use_global_groups TRUE, U in_group G, G name "committers"))))')
    rules = [
            (rgx('/vcreview/patches/all'),
             build_rset(all_patches_rql % final_patch_states_rql(), vid='vcreview.patches.table',
                        vtitle=all_patches_title)),
            (rgx('/vcreview/patches/worklist'),
             build_rset(user_worklist_rql, vid='vcreview.patches.table',
                        vtitle=user_worklist_title, setuser=True)),
    ]

class AllActivePatches(view.StartupView):
    __regid__ = 'vcreview.allactivepatches'
    replacement = 'vcreview/patches/all'

    def call(self):
        raise Redirect(self._cw.build_url(self.replacement), status=httplib.MOVED_PERMANENTLY)


class UserWorkList(AllActivePatches):
    __regid__ = 'vcreview.patches.worklist'
    replacement = 'vcreview/patches/worklist'


class PatchesTable(tableview.RsetTableView):
    __regid__ = 'vcreview.patches.table'
    paginable = True
    cellvids={1: 'vcreview.patch.authors', 2: 'vcreview.patch.reviewers', 3: 'vcreview.patch.committers'}
    # XXX selector

    def column_label(self, colidx, default, translate_func=None):
        if colidx == 1:
            return self._cw._('author')
        if colidx == 2:
            return self._cw._('patch_reviewer')
        if colidx == 3:
            return self._cw._('patch_committer')
        return super(PatchesTable, self).column_label(colidx, default, translate_func)


class FilteredPatchesTable(PatchesTable):
    __regid__ = 'vcreview.patches.table.filtered'
    layout_args = {'display_filter': 'top'}


class PatchAuthors(view.EntityView):
    __regid__ = 'vcreview.patch.authors'
    __select__ = is_instance('Patch')
    def entity_call(self, entity, **kwargs):
        self.w(u', '.join(set(xml_escape(rev.author) for rev in entity.patch_revision)))

class PatchReviewer(view.EntityView):
    __regid__ = 'vcreview.patch.reviewers'
    __select__ = is_instance('Patch')
    def entity_call(self, entity, **kwargs):
        self._cw.view('csv', entity.related('patch_reviewer'), 'null',
                      w=self.w)

class PatchCommitter(view.EntityView):
    __regid__ = 'vcreview.patch.committers'
    __select__ = is_instance('Patch')
    def entity_call(self, entity, **kwargs):
        self._cw.view('csv', entity.related('patch_committer'), 'null',
                      w=self.w)


class WorkListLink(component.CtxComponent):
    __regid__ = 'vcreview.worklist.warning'
    __select__ = (component.CtxComponent.__select__
                  & ~anonymous_user()
                  & rql_condition(
                      'P in_state PS, P patch_revision RE, RE from_repository R,'
                      'EXISTS(PS name "pending-review" AND P patch_reviewer U) '
                      'OR EXISTS(PS name IN ("reviewed", "deleted") AND (EXISTS(R repository_committer U) '
                      'OR EXISTS(U in_group G, G name "committers")))',
                      user_condition=True))
    context = 'left'
    order = -100

    def render(self, w, view=None):
        w(u'<table><tr><td>')
        w(tags.a(
            tags.img(src=self._cw.data_url('warning_icon.png'), alt=''),
            href=self._cw.build_url('vcreview/patches/worklist'),
            title=self._cw._('You have some patches in your worklist'),
            escapecontent=False)
          )
        w(u'</td><td>')
        w(tags.a(
            self._cw._('You have some patches in your worklist'),
            href=self._cw.build_url('vcreview/patches/worklist'))
          )
        w(u'</td></tr></table><br/>')



# copyright 2011-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-vcreview primary views and adapters for the web ui"""

__docformat__ = "restructuredtext en"
_ = unicode

from logilab.mtconverter import xml_escape

from cubicweb import tags
from cubicweb.predicates import score_entity, is_instance
from cubicweb.view import EntityView
from cubicweb.schema import display_name
from cubicweb.web.views import tabs, primary, ibreadcrumbs, uicfg

from cubes.vcsfile.views import primary as vcsfile
from cubes.vcreview.views import final_patch_states_rql
from cubes.vcreview.site_cubicweb import COMPONENT_CONTEXT

_pvs = uicfg.primaryview_section
_pvdc = uicfg.primaryview_display_ctrl
_abaa = uicfg.actionbox_appearsin_addmenu
_afs = uicfg.autoform_section


# patch primary view ###########################################################

_pvs.tag_subject_of(('Patch', 'patch_reviewer', '*'), 'attributes')
_pvs.tag_subject_of(('Patch', 'patch_committer', '*'), 'attributes')
_pvdc.tag_subject_of(('Patch', 'applied_at', '*'), {'vid': 'incontext'})

class PatchPrimaryView(tabs.TabbedPrimaryView):
    """Main Patch primary view"""
    __select__ = is_instance('Patch')

    tabs = [_('vcreview.patch.tab_main'),
            _('vcreview.patch.tab_head')]
    default_tab = 'vcreview.patch.tab_main'

    def render_entity_title(self, entity):
        self.w(u'<h1>%s <span class="state">[%s]</span></h1>'
               % (xml_escape(entity.dc_title()),
                  xml_escape(entity.cw_adapt_to('IWorkflowable').printable_state)))


class PatchPrimaryTab(tabs.PrimaryTab):
    """Summary tab for Patch"""
    __regid__ = 'vcreview.patch.tab_main'
    __select__ = is_instance('Patch')

    def render_entity_attributes(self, entity):
        super(PatchPrimaryTab, self).render_entity_attributes(entity)
        self.w(u'<h4>%s</h4>' % self._cw._('Patch revisions'))
        rset = self._cw.execute(
            'Any R,RA,RB,RC,RD,R ORDERBY RC DESC '
            'WHERE X eid %(x)s, X patch_revision R,'
            'R author RA, R branch RB, R creation_date RC, R description RD',
            {'x': entity.eid})
        _, __ = self._cw._, self._cw.__
        self.wview('table', rset,
                   headers=[__('Revision'), __('author'), __('branch'),
                            __('creation_date'), __('commit message')])

class PatchHeadTab(EntityView):
    """Revision view of the tip most version of the patch

    (with comment and task)"""
    __regid__ = 'vcreview.patch.tab_head'
    __select__ = is_instance('Patch')

    def entity_call(self, entity):
        tip = entity.tip()
        if tip:
            tip.view('primary', w=self.w)


#_pvs.tag_object_of(('*', 'patch_revision', '*'), 'hidden') # in breadcrumbs

# repository primary view ######################################################

_pvs.tag_subject_of(('Repository', 'repository_committer', '*'), 'attributes')
_pvs.tag_subject_of(('Repository', 'repository_reviewer', '*'), 'attributes')


class RepositoryPatchesTab(EntityView):
    __regid__ = _('vcreview.patches_tab')
    __select__ = is_instance('Repository') & score_entity(lambda x: x.has_review)

    def entity_call(self, entity):
        entity.view('vcreview.repository.patches', w=self.w)

vcsfile.RepositoryPrimaryView.tabs.append(RepositoryPatchesTab.__regid__)


class RepositoryPatchesTable(EntityView):
    __regid__ = 'vcreview.repository.patches'
    __select__ = is_instance('Repository') & score_entity(lambda x: x.has_review)

    rql = ('Any P,P,P,P,PB,PS ' # XXX need a sub query to find patch tip
           'GROUPBY P,PB,PS '
           'WHERE '
           '      RE branch PB, '
           '      P in_state PS, '
           '      P patch_revision RE, '
           '      RE from_repository R, '
           '      R eid %(x)s')

    def entity_call(self, entity):
        linktitle = self._cw._('Patches for %s') % entity.dc_title()
        linkurl = self._cw.build_url(rql=self.rql % {'x': entity.eid},
                                     vtitle=linktitle)
        self.w('<p>%s. <a href="%s">%s</a></p>' % (
            self._cw._('Table below only show active patches'),
            xml_escape(linkurl), self._cw._('Show all patches')))
        rql = self.rql + ', NOT PS name %s' % final_patch_states_rql()
        rset = self._cw.execute(rql, {'x': entity.eid})
        self.wview('vcreview.patches.table.filtered', rset, 'noresult')

# task primary view ############################################################

_pvs.tag_object_of(('*', 'has_activity', 'Task'), 'attributes')

class InsertionPointInContextView(EntityView):
    __regid__ = 'incontext'
    __select__ = is_instance('InsertionPoint')
    def entity_call(self, entity):
        purl = entity.parent.absolute_url()
        url = '%s#%s%s' % (purl, COMPONENT_CONTEXT, entity.eid)
        self.w(u'<a href="%s">%s</a>' % (url, entity.parent.dc_title()))


# breadcrumbs ##################################################################

class PatchIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance('Patch')

    def parent_entity(self):
        return self.entity.repository

class TaskBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = (is_instance('Task')
                  & score_entity(lambda x: x.reverse_has_activity))

    def parent_entity(self):
        return self.entity.reverse_has_activity[0]


class InsertionPointBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance('InsertionPoint')

    def parent_entity(self):
        return self.entity.parent

    def breadcrumbs(self, view=None, recurs=None):
        path = super(InsertionPointBreadCrumbsAdapter, self).breadcrumbs(
            view, recurs)
        return [x for x in path if getattr(x, 'eid', None) != self.entity.eid]



# copyright 2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-vcreview facets to filter patches"""

__docformat__ = "restructuredtext en"

from rql import nodes

from cubicweb.web import facet
from cubicweb.predicates import is_instance

class PatchRepositoryFacet(facet.RQLPathFacet):
    __regid__ = 'vcreview.patch_repository'
    __select__ = is_instance('Patch')
    path = ['X patch_revision RE', 'RE from_repository R', 'R title T']
    title = 'Repository'
    filter_variable = 'R'
    label_variable = 'T'


class PatchReviewerFacet(facet.RelationFacet):
    __regid__ = 'vcreview.patch_reviewer'
    rtype = 'patch_reviewer'
    role = 'subject'
    target_attr = 'login'


class PatchAuthorFacet(facet.RQLPathFacet):
    __regid__ = 'vcreview.patch_author'
    __select__ = is_instance('Patch')
    title = 'Author'
    path = ['X patch_revision RE', 'RE author A']
    filter_variable = 'A'

class PatchBranchFacet(facet.RQLPathFacet):
    __regid__ = 'vcreview.patch_branch'
    __select__ = is_instance('Patch')
    title = 'Branch'
    path = ['X patch_revision RE', 'RE branch B']
    filter_variable = 'B'

class PatchCommitterFacet(facet.RelationFacet):
    __regid__ = 'vcreview.patch_committer'
    rtype = 'patch_committer'
    role = 'subject'
    target_attr = 'login'


class PatchHasTodoTask(facet.HasRelationFacet):
    """Filter out patches with todo tasks"""
    __regid__ = 'vcreview.open_task'
    __select__ = is_instance('Patch')
    title = _('has open task')

    def add_rql_restrictions(self):
        value = self._cw.form.get(self.__regid__)
        if not value: # no value sent for this facet
            return
        exists_node1 = nodes.Exists()
        exists_node2 = nodes.Exists()
        self.select.add_restriction(nodes.Or(exists_node1, exists_node2))
        self.fill_exists(exists_node1,
                         [('has_activity', 'subject'), ('in_state', 'subject')])
        self.fill_exists(exists_node2,
                         [('patch_revision', 'subject'), ('point_of', 'object'),
                          ('has_activity', 'subject'), ('in_state', 'subject')])

    def fill_exists(self, node, rtypes_roles):
        """Fill a node with relation restrictions from a sequence of rtypes
        and roles
        """
        pvar = self.filtered_variable
        for rtype, role in rtypes_roles:
            var = self.select.make_variable()
            if role == 'subject':
                node.add_relation(pvar, rtype, var)
            else:
                node.add_relation(var, rtype, pvar)
            pvar = var
        node.add_constant_restriction(pvar, 'name', 'todo', 'String')



# copyright 2011-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from __future__ import with_statement

import os.path as osp
import functools
import tempfile
import tarfile
from shutil import rmtree

from logilab.common.decorators import classproperty

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.web import facet

from cubes.vcsfile.testutils import init_vcsrepo, HGRCMixin
from cubes.vcreview.views.facets import PatchHasTodoTask


class PatchReviewWorkflowTC(HGRCMixin, CubicWebTC):

    @classmethod
    def setUpClass(cls):
        cls.__tmpdir = tempfile.mkdtemp(prefix='vcreview-test-')
        tarfile.open(cls.repo_path + '.tar').extractall(cls.__tmpdir)

    @classmethod
    def tearDownClass(cls):
        cls.cleanup()

    @classmethod
    def cleanup(cls):
        try:
            rmtree(cls.__tmpdir)
        except:
            pass

    @classproperty
    def full_repo_path(cls):
        return osp.join(cls.__tmpdir, u'repo')

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            # create users
            create_user = functools.partial(self.create_user, cnx)
            self.reviewer1 = create_user('Victor', ('users', 'reviewers'),
                                         email=u'victor@ratax.es').eid
            self.reviewer2 = create_user('Podular', ('users', 'reviewers'),
                                         email=u'podular@elephan.tz').eid
            self.committer = create_user('committer', ('users',)).eid
            cnx.commit()

            # import revisions from repository
            with cnx.security_enabled(write=False):
                with cnx.deny_all_hooks_but('metadata'):
                    cnx.create_entity('Repository', type=u'mercurial',
                                      encoding=u'utf8',
                                      local_cache=self.full_repo_path,
                                      repository_committer=self.committer,
                                      has_review=True)
                    cnx.commit()
            init_vcsrepo(self.repo)

            # ask review
            rset = cnx.execute('Patch P ORDERBY P DESC '
                               'WHERE P in_state S, '
                               '      S name "in-progress"')
            for patch in rset.entities():
                patch.cw_adapt_to('IWorkflowable').fire_transition('ask review')
            cnx.commit()

            # record eids of patches pending review
            rset = cnx.execute('Patch P ORDERBY P DESC '
                               'WHERE P in_state S, '
                               '      S name "pending-review"')
            self.patch1eid = rset[2][0]
            self.patch2eid = rset[3][0]
            cnx.commit()

            # create task
            self.task1eid = cnx.create_entity(
                'Task', reverse_has_activity=self.patch1eid, title=u'todo 1').eid
            cnx.commit()

    def test_wf(self):
        with self.admin_access.client_cnx() as cnx:
            patch = cnx.entity_from_eid(self.patch1eid)
            self.assertEqual(patch.patch_reviewer[0].eid, self.reviewer2)
        with self.new_access('Victor').client_cnx() as cnx:
            rset = cnx.execute('Any X WHERE X eid %(x)s', {'x': self.patch1eid})
            self.assertListEqual(sorted(x[0] for x in self.action_submenu(cnx, rset, 'workflow')),
                                 ['accept', 'ask rework', 'fold', 'view history', 'view workflow'])
            # reviewer1 submit its second patch for review: while reviewer2
            # already has a patch for review, it should be picked anyway as a
            # patch creator can't be its reviewer
            patch = cnx.execute('Any X WHERE X eid %(x)s', {'x': self.patch2eid}).get_entity(0, 0)
            self.assertEqual(patch.patch_reviewer[0].eid, self.reviewer2)
            # create a task and mark the task1 as done
            task2eid = cnx.create_entity(
                'Task', reverse_has_activity=patch, title=u'todo 2').eid
            task = cnx.execute('Any X WHERE X eid %(x)s', {'x': self.task1eid}).get_entity(0, 0)
            task.cw_adapt_to('IWorkflowable').fire_transition('done')
            cnx.commit()
        with self.new_access('Podular').client_cnx() as cnx:
            rset = cnx.execute('Any X WHERE X eid %(x)s', {'x': self.patch1eid})
            self.assertListEqual(sorted(x[0] for x in self.action_submenu(cnx, rset, 'workflow')),
                                 ['accept', 'ask rework', 'fold', 'view history', 'view workflow'])
            patch = rset.get_entity(0, 0)
            patch.cw_adapt_to('IWorkflowable').fire_transition('accept')
            cnx.commit()
            patch.cw_clear_all_caches()
            self.assertListEqual(sorted(x[0] for x in self.action_submenu(cnx, rset, 'workflow')),
                                 [u'ask review', u'ask rework', u'fold', u'view history', u'view workflow'])
            # create a task and mark task2 as done
            task3eid = cnx.create_entity(
                'Task', reverse_has_activity=patch, title=u'todo 3').eid
            task = cnx.execute('Any X WHERE X eid %(x)s', {'x': task2eid}).get_entity(0, 0)
            task.cw_adapt_to('IWorkflowable').fire_transition('done')
            cnx.commit()
        with self.new_access('committer').client_cnx() as cnx:
            rset = cnx.execute('Any X WHERE X eid %(x)s', {'x': self.patch1eid})
            self.assertListEqual(sorted(x[0] for x in self.action_submenu(cnx, rset, 'workflow')),
                                 ['apply', 'ask review', 'ask rework', 'fold', 'reject', 'view history', 'view workflow'])
            patch = rset.get_entity(0, 0)
            patch.cw_adapt_to('IWorkflowable').fire_transition('apply')
            cnx.commit()
            patch.cw_clear_all_caches()
            self.assertListEqual(sorted(x[0] for x in self.action_submenu(cnx, rset, 'workflow')),
                                 ['view history'])
            # mark task3 as done
            task = cnx.execute('Any X WHERE X eid %(x)s', {'x': task3eid}).get_entity(0, 0)
            task.cw_adapt_to('IWorkflowable').fire_transition('done')
            cnx.commit()


    def test_review_adapter_reviewers_rset(self):
        """check reviewer configuration

        1) Test that explicit `repository_reviewer` are taken in account
        2) Test that `reviewer` group can be disabled on a repo.
        """
        with self.admin_access.client_cnx() as cnx:
            otherrepo = cnx.create_entity('Repository', type=u'mercurial',
                                          encoding=u'utf8',
                                          local_cache=u'/dev/null',
                                          has_review=True)
            cnx.execute('SET X repository_reviewer U WHERE U login "committer"')
            patch = cnx.entity_from_eid(self.patch1eid)
            ireview = patch.cw_adapt_to('IPatchReviewControl')
            self.assertEqual(set(x for x,_ in ireview.reviewers_rset().rows),
                             set([self.reviewer2,
                                  self.committer]))
            cnx.execute('SET X use_global_groups FALSE WHERE NOT X eid %(e)s', {'e': otherrepo.eid})
            patch.cw_clear_all_caches()
            ireview = patch.cw_adapt_to('IPatchReviewControl')
            self.assertEqual(set(x for x,_ in ireview.reviewers_rset().rows),
                             set([self.committer]))


PatchReviewWorkflowTC.repo_path = osp.join(PatchReviewWorkflowTC.datadir,
                                      u'demo-repo-final')


class VCReviewFacetTC(CubicWebTC):

    def test_patch_has_open_task_facet(self):
        with self.admin_access.web_request() as req:
            req.create_entity('Patch', patch_name=u'turlu')
            rset = req.cnx.execute('Patch X')
            rqlst = rset.syntax_tree().copy()
            select = rqlst.children[0]
            filtered_variable, baserql = facet.init_facets(rset, select)

            f = PatchHasTodoTask(req, rset=rset, select=select,
                                 filtered_variable=filtered_variable)
            req.form[f.__regid__] = "1"
            f.add_rql_restrictions()
            self.assertEqual(f.select.as_string(),
                             "DISTINCT Any  WHERE X is Patch, (EXISTS(X has_activity A, "
                             "A in_state B, B name 'todo')) OR (EXISTS(X patch_revision C, "
                             "D point_of C, D has_activity E, E in_state F, F name 'todo'))")


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.


from cubes.vcsfile.docparser import Diff2HTMLTransform
from cubicweb.cwconfig import register_persistent_options

register_persistent_options( (
    # user config
    ('autoreview',
     {'type' : 'yn', 'default': False,
      'help': _('Automatically set new patches in the pending-review state'),
      'group': 'vcreview',
      }),
    ))

COMPONENT_CONTEXT = 'navcontentbottom'
def insert_point_callback(ipid, trdata, w):
    context = trdata.appobject
    cw = context._cw
    e_schema = getattr(context, 'e_schema', None)
    if e_schema is None:
        # not an entity
        return
    if 'point_of' not in e_schema.objrels:
        # no InsertionPoint on this entity
        return
    ipoint = cw.execute('InsertionPoint IP WHERE IP point_of X, X eid %(x)s, IP lid %(id)s',
                        {'x': context.eid, 'id': ipid})
    if ipoint:
        assert len(ipoint) == 1
        ipoint = ipoint.get_entity(0, 0)
        extra = {}
    else:
        ipoint = cw.vreg['etypes'].etype_class('InsertionPoint')(cw)
        ipoint.eid = cw.varmaker.next()
        extra = {'formparams': {'lid': ipid, 'point_of': context.eid}}
    w(u'<div id="%s%s">' % (COMPONENT_CONTEXT, ipoint.eid))
    for comp in cw.vreg['ctxcomponents'].poss_visible_objects(
        cw, rset=ipoint.cw_rset, entity=ipoint, context=COMPONENT_CONTEXT,
        __cache=None, **extra):
        if comp is not None:
            comp.render(w=w)
    w(u'</div>')

def register_transforms(engine, verb=True):
    if Diff2HTMLTransform.name in engine.transforms:
        engine.remove_transform(Diff2HTMLTransform.name, 'text/x-diff')
    engine.add_transform(Diff2HTMLTransform(insert_point_callback))

from cubicweb.mttransforms import ENGINE
register_transforms(ENGINE)




"""cubicweb-vcreview application package

patch review system on top of vcsfile
"""

import re

EMAIL_RE = re.compile('<([^>]*@[^>]*)>')
def users_by_author(execute, author):
    """Find CWUser related to a patch

    We use revision author email to search for related CWUser."""
    match = EMAIL_RE.search(author)
    if match is not None:
        author_email = match.group(1)
    elif '@' in author:
        author_email = author
    else:
        return []
    rset = execute('DISTINCT Any X WHERE X is CWUser, X use_email E, '
                            'E address ILIKE %(email)s', {'email': author_email})
    return [ueid for ueid, in rset]

def _prepare_patch(patch):
    """Drop all information but the username and patch"""
    cleanpatch = []
    for line in patch.splitlines():
        if line.startswith('# User') or not line.startswith('#'):
            if line.startswith('@@'):
                line = '@@\n'
            cleanpatch.append(line)
    return cleanpatch


def compare_patches(old, new):
    """Compare two patches in git format."""
    return _prepare_patch(old) == _prepare_patch(new)



pwf = get_workflow_for('Patch', ask_confirm=False)
delete = pwf.transition_by_name('file deleted')
reviewed = pwf.state_by_name('reviewed')
rql('SET S allowed_transition T WHERE S eid %(s)s, T eid %(t)s, NOT S allowed_transition T',
    {'s': reviewed.eid, 't': delete.eid})



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from logilab.common.testlib import TestCase, unittest_main, mock_object
from logilab.mtconverter import TransformData

from cubicweb import devtools
from cubicweb.mttransforms import ENGINE

from cubes.vcsfile.docparser import Diff2HTMLTransform

class FakeRequest(object):
    def add_css(self, file): pass

class DiffParserTC(TestCase):
    @classmethod
    def setUpClass(cls):
        # in case a previous test set this up
        try:
            ENGINE.remove_transform(Diff2HTMLTransform.name)
        except KeyError:
            pass

    @classmethod
    def tearDownClass(cls):
        # reset to defaults
        from cubes.vcreview.site_cubicweb import register_transforms
        register_transforms(ENGINE)

    def transform(self, file):
        data = open(self.datapath(file)).read()
        trdata = TransformData(data, 'text/x-diff', 'utf-8')
        trdata.appobject = mock_object(_cw=FakeRequest())
        return ENGINE.convert(trdata, 'text/annotated-html').data

    def setUp(self):
        def insert_point_cb(ipid, trdata, w):
            w('<hr class="%s"/>' % ipid)
            self.ipid = ipid
        ENGINE.add_transform(Diff2HTMLTransform(insert_point_cb))

    def tearDown(self):
        ENGINE.remove_transform(Diff2HTMLTransform.name)

    def test_insert_point_1(self):
        htmldata = self.transform('pytestgc.diff')
        self.assertEqual(self.ipid, 5)
        self.assertTrue(htmldata.startswith('<div class="text/x-diff">'))

    def test_insert_point_2(self):
        htmldata = self.transform('nonregr1.diff')
        self.assertEqual(self.ipid, 2)
        self.assertTrue(htmldata.startswith('<div class="text/x-diff">'))

    def test_insert_point_3(self):
        htmldata = self.transform('nonregr2.diff')
        self.assertEqual(self.ipid, 3)
        self.assertTrue(htmldata.startswith('<div class="text/x-diff">'))

if __name__ == '__main__':
    unittest_main()



# copyright 2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-vcreview automatic tests"""

import os.path as osp
from shutil import rmtree
import tempfile
import tarfile

from logilab.common.decorators import classproperty


from cubicweb.devtools import testlib

from cubes.vcsfile.testutils import init_vcsrepo, HGRCMixin

class AutomaticWebTest(HGRCMixin, testlib.AutomaticWebTest):
    no_auto_populate = ('Repository',
                        'Revision',
                        'Patch',
                        'Comment',
                        'InsertionPoint',)
    ignored_relations = set(('parent_revision',
                             'from_repository',
                             'patch_revision',
                             'nosy_list',
                             'comments',
                             'depends_on',
                             'point_of'))

    def to_test_etypes(self):
        return set(('Repository',
                    'Revision',
                    'Patch',
                    'InsertionPoint',
                    ))


    @classmethod
    def setUpClass(cls):
        cls.__tmpdir = tempfile.mkdtemp(prefix='vcreview-test-')
        tarfile.open(cls.repo_path + '.tar').extractall(cls.__tmpdir)

    @classmethod
    def tearDownClass(cls):
        cls.cleanup()

    @classmethod
    def cleanup(cls):
        try:
            rmtree(cls.__tmpdir)
        except:
            pass

    @classproperty
    def full_repo_path(cls):
        return osp.join(cls.__tmpdir, u'repo')

    # XXX presetup
    def custom_populate(self, how_many, cnx):
        self.create_user(cnx, 'Babar', ('users', 'reviewers', 'committers'),
                         email=u'babar@jungle.net')
        self.create_user(cnx, 'Victor', ('users', 'reviewers'),
                         email=u'victor@ratax.es')
        cnx.commit()
        cnx.create_entity('Repository',
                          type=u'mercurial',
                          encoding=u'utf8',
                          local_cache=self.full_repo_path,
                          has_review=True)
        cnx.commit()
        init_vcsrepo(self.repo)

AutomaticWebTest.repo_path = osp.join(AutomaticWebTest.datadir,
                                      u'demo-repo-final')

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()




from cubicweb.predicates import is_instance, is_in_state
from cubicweb.sobjects.notification import StatusChangeMixIn, NotificationView
from cubes.nosylist.sobjects import NosyListRecipientsFinder
from cubes.task import hooks as task
from cubes.vcreview.site_cubicweb import COMPONENT_CONTEXT

MAX_PATCH_SIZE = 100 << 10 # 100 Ko


class NosyListPatchRecipientsFinder(NosyListRecipientsFinder):
    """Recipient finder of Patches.

    Merge users in the patch nosy list with those of the associated repository.

    This recipient finder is implemented instead of using a nosy list
    propagation rule to avoid useless (and costly) propagations when a user
    subscribes to a project with a "big" set of patches.
    """
    __select__ = is_instance('Patch')

    def recipients(self):
        recipients = set(super(NosyListPatchRecipientsFinder, self).recipients())
        patch_eid = self.cw_rset[self.cw_row or 0][self.cw_col or 0]
        rset = self._cw.execute('Any X WHERE P eid %(p)s, P patch_revision R,'
                                ' R from_repository X', {'p': patch_eid})
        repo_finder = self._cw.vreg['components'].select(
            'recipients_finder', self._cw, rset=rset, row=0, col=0)
        recipients.update(set(repo_finder.recipients()))
        return list(recipients)


class PatchAddedView(NotificationView):
    """Notification view for new patches."""
    __regid__ = 'notif_after_add_entity'
    __select__ = is_instance('Patch')
    msgid_timestamp = False
    content = """url: %(url)s

%(content)s
"""

    def subject(self):
        entity = self.cw_rset.get_entity(0, 0)
        return self._cw._(u'[%(repo)s] new patch: %(patch)s') % {
            'repo': entity.repository.dc_title(),
            'patch': entity.dc_title()}

    def context(self, **kwargs):
        entity = self.cw_rset.get_entity(0, 0)
        patch = entity.email_export(size=MAX_PATCH_SIZE)
        return super(PatchAddedView, self).context(content=patch, **kwargs)


class PatchStatusChangeView(StatusChangeMixIn, NotificationView):
    __select__ = is_instance('Patch') & is_in_state('pending-review',
                                                    'in-progress',
                                                    'reviewed',
                                                    'applied',
                                                    'rejected')

    content = StatusChangeMixIn.content + '''
%(reviewer)s

%(patch)s
'''

    def subject(self):
        entity = self.cw_rset.get_entity(0, 0)
        return self._cw._(u'[%(repo)s] patch %(state)s: %(patch)s') % {
            'repo': entity.repository.dc_title(),
            'patch': entity.dc_title(),
            'state': self._cw.__(self._kwargs['current_state'])}

    def context(self, **kwargs):
        context = super(PatchStatusChangeView, self).context(**kwargs)
        entity = self.cw_rset.get_entity(0, 0)
        tasks = []
        for task in self._cw.execute(
            u'Any T,TT ORDERBY TCD DESC '
            'WHERE T in_state TS, TS name "todo" '
            'WITH T,TT,TCD BEING ('
            '(Any T,TT,TCD WHERE '
            'X has_activity T, T title TT, T creation_date TCD, X eid %(x)s)'
            ' UNION '
            '(Any T,TT,TCD WHERE '
            'X patch_revision RE, IP point_of RE, IP has_activity T, '
            'T title TT, T creation_date TCD, X eid %(x)s)'
            ')', {'x': entity.eid}).entities():
            tasks.append(task.dc_title())
        if tasks:
            context['comment'] += '\n\n%s\n\n- %s' % (
                self._cw._('remaining tasks:'), '\n- '.join(tasks))
        else:
            context['comment'] += '\n\n%s' % self._cw._('(No recorded task)')

        context['patch'] = ''
        if self._kwargs['current_state'] == 'pending-review':
            patch = entity.email_export(size=MAX_PATCH_SIZE)
            context['patch'] = patch

        if entity.patch_reviewer:
            context['reviewer'] = self._cw._('reviewer: ') + ', '.join([user.dc_title() for user in entity.patch_reviewer])
        else:
            context['reviewer'] = self._cw._('reviewer: ')

        return context


class TaskAddedView(task.TaskAddedView):
    """get notified from new tasks"""
    __select__ = is_instance('Task')
    content_attr = 'description'

    def subject(self):
        entity = self.cw_rset.get_entity(0, 0)
        patch = entity.activity_of
        if not patch:
            return super(TaskAddedView, self).subject()
        if patch.__regid__ == 'InsertionPoint':
            patch = patch.parent.reverse_patch_revision
            if not patch:
                return super(TaskAddedView, self).subject()
            patch = patch[0]
        return self._cw._(u'[%(repo)s] patch task: %(patch)s - %(task)s') % {
            'repo': patch.repository.dc_title(),
            'patch': patch.dc_title(),
            'task': entity.dc_title()}

    def context(self, **kwargs):
        kwargs = super(TaskAddedView, self).context(**kwargs)
        task = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        concerned = task.activity_of
        if concerned:
            # build a URL for the task either in the context of an
            # InsertionPoint or directly using the related Patch.
            if concerned.cw_etype == 'InsertionPoint':
                revision = concerned.parent
                fragment = '#{ctx}{eid}'.format(ctx=COMPONENT_CONTEXT,
                                                eid=concerned.eid)
                url = revision.absolute_url(__secure__=True) + fragment
            else:
                # This should be a Patch.
                url = concerned.absolute_url(__secure__=True)
            kwargs['url'] = url
        return kwargs


class ReviewerSetNotificationView(NotificationView):
    __regid__ = 'vcreview.notifications.reviewer-set'
    __select__ = is_instance('Patch')
    content = _('''
You have been assigned for review of patch %(title)s

url: %(url)s

%(patch)s
''')
    def recipients(self):
        reviewer = self.cw_extra_kwargs['user']
        email = reviewer.cw_adapt_to('IEmailable').get_email()
        if email:
            return [(email, reviewer.property_value('ui.language'))]
        return []

    def subject(self):
        patch = self.cw_rset.get_entity(0, 0)
        return self._cw._('[%(repo)s] patch assigned: %(patch)s') % {
            'app': self._cw.vreg.config.appid,
            'repo': patch.repository.dc_title(),
            'patch': patch.dc_title()}

    def context(self, **kwargs):
        context = super(ReviewerSetNotificationView, self).context(**kwargs)
        entity = self.cw_rset.get_entity(0, 0)
        patch = entity.email_export(size=MAX_PATCH_SIZE)
        context['patch'] = patch
        return context


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, (TaskAddedView,))
    vreg.register_and_replace(TaskAddedView, task.TaskAddedView)



pwf = rql('Workflow X WHERE X name "patch workflow"').get_entity(0, 0)
inprogress = pwf.state_by_name('in-progress')
deleted = pwf.state_by_name('deleted')
pwf.add_transition(_('reopen'), deleted, inprogress,
                   ('managers',))


add_attribute('Patch', 'branch')
add_attribute('Patch', 'originator')

for patch in rql('Patch X').entities():
    revision = patch.revisions[0].rev
    patch.set_attributes(branch=revision.branch, originator=revision.author)


add_relation_definition('CWUser', 'interested_in', 'Patch')
add_relation_definition('VersionContent', 'nosy_list', 'CWUser')
add_relation_definition('DeletedVersionContent', 'nosy_list', 'CWUser')
add_relation_definition('InsertionPoint', 'nosy_list', 'CWUser')



# -*- coding: utf-8 -*-

# copyright 2011-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import os
from os import path as osp
from shutil import rmtree
import re
import tempfile
import tarfile
from subprocess import check_call, Popen, PIPE
import hglib

from logilab.common.decorators import classproperty

from cubicweb.devtools.testlib import MAILBOX, CubicWebTC

from cubes.vcsfile import bridge
from cubes.vcsfile.testutils import init_vcsrepo, HGRCMixin, VCSFileTC
from cubes.vcreview.testutils import HgHelperTC

def revnums(patch):
    return sorted(vc.revision for vc in patch.patch_revision)


class AutoReviewPrefTC(VCSFileTC):
    _repo_path = u'newrepo'

    @classmethod
    def pre_setup_database(cls, cnx, config):
        super(AutoReviewPrefTC, cls).pre_setup_database(cnx, config)
        repo = cnx.entity_from_eid(cls.vcsrepoeid)
        repo.cw_set(has_review=True)
        babar = cls.create_user(cnx, 'Babar', ('users', 'reviewers', 'committers'),
                                email=u'babar@ratax.es')
        victor = cls.create_user(cnx, 'Victor', ('users', 'reviewers'),
                                 email=u'victor@ratax.es')
        arthur = cls.create_user(cnx, 'Arthur', ('users', 'reviewers'),
                                 email=u'arthur@ratax.es')
        cnx.create_entity('CWProperty', pkey=u"vcreview.autoreview", value=u"1",
                          for_user=babar)
        cnx.create_entity('CWProperty', pkey=u"vcreview.autoreview", value=u"0",
                          for_user=victor)
        cnx.commit()

    def test_user_pref(self):
        self.refresh()
        with hglib.open(self.repo_path) as hgrepo:
            root = hgrepo.root()
            for user in 'arthur', 'babar', 'victor':
                fname = osp.join(root, user)
                with open(fname, 'w') as f:
                    f.write(user)
                hgrepo.add(fname)
                hgrepo.commit(message="commit from %s" % user,
                              user="%s <%s@ratax.es>" % (user, user))
        with self.repo.internal_cnx() as cnx:
            repo = cnx.entity_from_eid(self.vcsrepoeid)
            repo.cw_adapt_to('VCSRepo').import_content(commitevery=1)
        # ensure patches exists and are at correct WF status
        ref = {'babar <babar@ratax.es>': 'pending-review',
               'victor <victor@ratax.es>': 'in-progress',
               'arthur <arthur@ratax.es>': 'in-progress',
               }
        with self.admin_access.client_cnx() as cnx:
            rset = cnx.execute('Any A, S WHERE P patch_revision REV, '
                               'REV author A, P in_state ST, ST name S')
            for user, status in rset: # pop to check "one commit per user"
                self.assertEqual(ref.pop(user), status, "%s: got %s" % (user, status))

class PatchLifeCycleTC(HgHelperTC):

    def test_root_publishing(self):

        hg = self.hgrepo
        a = self.hg_append_to

        ### 1) root initially draft becomes public
        #############################################################

        # create a simple root changeset as draft
        a('jungle', 'Babar')
        hg.commit(message='Initial jungle content')
        self.refresh()

        # We should have a new Patch linked to revision 0
        with self.admin_access.client_cnx() as cnx:
            revs = self.all_revs(cnx)
            self.assertEqual(1, len(revs))
            self.assertTrue(revs[0].reverse_patch_revision)
            patch = revs[0].reverse_patch_revision[0]
            self.assertEqual('Initial jungle content', patch.patch_name)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('in-progress', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual([], wf_history)

        # turn the changeset public
        hg.phase(['0'], public=True)
        self.refresh()
        # check that the patch has been "applied"
        with self.admin_access.client_cnx() as cnx:
            patch = cnx.entity_from_eid(patch.eid)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('applied', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual(['applied'], wf_history)

    def test_draft_publishing(self):
        self.test_root_publishing()
        hg = self.hgrepo
        a = self.hg_append_to

        with self.admin_access.client_cnx() as cnx:
            oldpatch = self.all_revs(cnx)[0].reverse_patch_revision[0].eid

        ### 2) draft changeset becomes public
        #############################################################

        # create a draft child
        a('jungle', 'Celestine')
        hg.commit(message='couple of elephants')
        self.refresh()

        # We should have a new Patch linked to revision 1
        with self.admin_access.client_cnx() as cnx:
            for rev in self.all_revs(cnx):
                if rev.description == 'couple of elephants':
                    break
            else:
                self.fail('missing draft child')
            self.assertTrue(rev.reverse_patch_revision)
            patch = rev.reverse_patch_revision[0]
            self.assertNotEqual(oldpatch, patch.eid)
            self.assertEqual('couple of elephants', patch.patch_name)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('in-progress', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual([], wf_history)
            # old patch is unchanged
            oldpatch_state = cnx.entity_from_eid(oldpatch).cw_adapt_to('IWorkflowable').state
            self.assertEqual('applied', oldpatch_state)

        # make the child public
        hg.phase(['1'], public=True)
        self.refresh()

        # check that the patch has been "applied"
        with self.admin_access.client_cnx() as cnx:
            patch = cnx.entity_from_eid(patch.eid)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('applied', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual(['applied'], wf_history)

    def test_empty_desc(self):
        self.test_root_publishing()
        hg = self.hgrepo
        a = self.hg_append_to

        a('jungle', 'Pam')
        hg.commit(message=' ') # sic
        self.refresh()

    def test_simple_amend(self):
        self.test_root_publishing()
        hg = self.hgrepo
        a = self.hg_append_to
        with self.admin_access.client_cnx() as cnx:
            oldpatch = self.all_revs(cnx)[0].reverse_patch_revision[0].eid

        ### 3) simple amend that ends up published
        #############################################################

        # add a new draft children again
        a('jungle', 'Pam')
        hg.commit(message='XXX')
        self.refresh()

        # We should have a new Patch linked to revision 1
        with self.admin_access.client_cnx() as cnx:
            rev = self.all_revs(cnx)[-1]
            assert rev.description == 'XXX'
            self.assertTrue(rev.reverse_patch_revision)
            patch = rev.reverse_patch_revision[0]
            self.assertNotEqual(oldpatch, patch.eid)
            self.assertEqual('XXX', patch.patch_name)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('in-progress', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual([], wf_history)

        # Fix changeset content and description
        path = osp.join(self.tmprepo, 'jungle')
        with open(path, 'r') as f:
            data = f.read().replace('Pam', 'Pom')
        with open(path, 'w') as f:
            f.write(data)
        self.hg('commit', '--amend', '-m', 'first child')
        self.refresh()
        # test that the new revision is added to the same patch
        with self.admin_access.client_cnx() as cnx:
            patch = cnx.entity_from_eid(patch.eid)
            rev = self.all_revs(cnx)[-1]
            assert rev.description == 'first child'
            newpatch = rev.reverse_patch_revision[0]
            self.assertEqual(patch.eid, newpatch.eid)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('in-progress', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual([], wf_history)
            self.assertEqual('first child', patch.patch_name)

        # yet another successor. A public one
        self.hg('commit', '--amend', '-m', 'First child')
        hg.phase(['.'], public=True)
        self.refresh()
        # check that the patch has been "applied"
        with self.admin_access.client_cnx() as cnx:
            patch = cnx.entity_from_eid(patch.eid)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('applied', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual(['applied'], wf_history)
            self.assertEqual(3, len(patch.patch_revision))


    def test_prune(self):
        self.test_root_publishing()
        hg = self.hgrepo
        a = self.hg_append_to
        with self.admin_access.client_cnx() as cnx:
            oldpatch = self.all_revs(cnx)[0].reverse_patch_revision[0].eid

        ### 4) simple amend that ends up pruned
        #############################################################

        # add a new draft child again
        a('jungle', 'Zephir')
        hg.commit('seocnd child')  # sic
        self.refresh()

        # look for a new patch
        with self.admin_access.client_cnx() as cnx:
            rev = self.all_revs(cnx)[-1]
            assert rev.description == 'seocnd child'
            self.assertTrue(rev.reverse_patch_revision)
            patch = rev.reverse_patch_revision[0]
            self.assertNotEqual(oldpatch, patch.eid)
            self.assertEqual('seocnd child', patch.patch_name)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('in-progress', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual([], wf_history)

        # amend the patch
        self.hg('commit', '--amend', '-m', 'second child')
        self.refresh()
        # test that patch has two revisions
        with self.admin_access.client_cnx() as cnx:
            patch = cnx.entity_from_eid(patch.eid)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('in-progress', patch_state)
            self.assertEqual(2, len(patch.patch_revision))
            wf_history = self.get_wf_history(patch)
            self.assertEqual([], wf_history)

        prec = self.get_id('.')
        hg.update('.^')
        self.hg('debugobsolete', prec)  # Zephir is a monkey, donkey!
        self.refresh()
        # patch should turn rejected
        with self.admin_access.client_cnx() as cnx:
            patch = cnx.entity_from_eid(patch.eid)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('rejected', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual(['rejected'], wf_history)
            self.assertEqual(2, len(patch.patch_revision))

    def test_zombie(self):
        self.test_root_publishing()
        hg = self.hgrepo
        a = self.hg_append_to
        with self.admin_access.client_cnx() as cnx:
            oldpatch = self.all_revs(cnx)[0].reverse_patch_revision[0].eid

        ### 5) changesets turning zombie at some point
        #############################################################

        # create a new draft child
        a('jungle', 'Flare')
        hg.commit('second child')
        self.refresh()

        # look for a new patch
        with self.admin_access.client_cnx() as cnx:
            rev = self.all_revs(cnx)[-1]
            assert rev.description == 'second child'
            self.assertTrue(rev.reverse_patch_revision)
            patch = rev.reverse_patch_revision[0]
            self.assertNotEqual(oldpatch, patch.eid)
            self.assertEqual('second child', patch.patch_name)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('in-progress', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual([], wf_history)

        # have a successor out of the repo
        path = osp.join(self.tmprepo, 'jungle')
        with open(path, 'r') as f:
            data = f.read().replace('Flare', 'Flore')
        with open(path, 'w') as f:
            f.write(data)
        bundlepath = osp.join(self.tmprepo, 'flore.hg')
        self.hg('commit', '--amend', '-m', 'second child')
        hg.bundle(bundlepath, rev='.', base='.^')
        self.hg('strip', '.')
        self.refresh()
        # patch should be "outdated"
        with self.admin_access.client_cnx() as cnx:
            patch = cnx.entity_from_eid(patch.eid)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('outdated', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual(['outdated'], wf_history)
            self.assertEqual(1, len(patch.patch_revision))

        # make the successor visible again
        self.hg('unbundle', bundlepath)
        self.refresh()
        # test should be alive again
        with self.admin_access.client_cnx() as cnx:
            patch = cnx.entity_from_eid(patch.eid)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('in-progress', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual(['outdated', 'in-progress'],
                             wf_history)
            self.assertEqual(2, len(patch.patch_revision))

    def test_delete_revision(self):
        self.test_root_publishing()
        hg = self.hgrepo
        a = self.hg_append_to

        # create a new draft child
        a('jungle', 'Flare')
        hg.commit('second child')
        self.refresh()

        with self.admin_access.client_cnx() as cnx:
            rev = self.all_revs(cnx)[-1]
            assert rev.description == 'second child'
            self.assertTrue(rev.reverse_patch_revision)
            eid_patch = rev.reverse_patch_revision[0].eid
        with self.repo.internal_cnx() as cnx:
            self.assertTrue(cnx.execute('DELETE Revision R WHERE R eid %(eid)s',
                                        {'eid': rev.eid}))
            cnx.commit()
        with self.admin_access.client_cnx() as cnx:
            self.assertFalse(cnx.execute('Patch P WHERE P eid %(eid)s',
                                          {'eid': eid_patch}))

    def test_split(self):
        self.test_root_publishing()
        hg = self.hgrepo
        a = self.hg_append_to
        with self.admin_access.client_cnx() as cnx:
            oldpatch = self.all_revs(cnx)[0].reverse_patch_revision[0].eid

        ### 6) split
        #############################################################

        # create a new draft changeset with two changes
        a('jungle', 'Alexandre')
        a('jungle', 'Isabelle')
        hg.commit('Moar children')
        self.refresh()

        # search for a new patch
        with self.admin_access.client_cnx() as cnx:
            rev = self.all_revs(cnx)[-1]
            assert rev.description == 'Moar children'
            self.assertTrue(rev.reverse_patch_revision)
            patch = rev.reverse_patch_revision[0]
            self.assertNotEqual(oldpatch, patch.eid)
            self.assertEqual('Moar children', patch.patch_name)
            patch_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('in-progress', patch_state)
            wf_history = self.get_wf_history(patch)
            self.assertEqual([], wf_history)

        # split them in two
        hg.update('.^')
        a('jungle', 'Alexandre')
        hg.commit('third child')
        a('jungle', 'Isabelle')
        hg.commit('fourth child')
        prec = self.get_id('-3')
        suc1 = self.get_id('-2')
        suc2 = self.get_id('-1')
        self.hg('debugobsolete', prec, suc1, suc2)
        self.refresh()
        # we should now have two patches
        with self.admin_access.client_cnx() as cnx:
            suc1 = self.all_revs(cnx)[-2]
            suc2 = self.all_revs(cnx)[-1]
            assert suc1.description == 'third child'
            assert suc2.description == 'fourth child'
            self.assertTrue(suc1.reverse_patch_revision)
            self.assertTrue(suc2.reverse_patch_revision)
            patch1 = suc1.reverse_patch_revision[0]
            patch2 = suc2.reverse_patch_revision[0]
            self.assertEqual(patch.eid, patch1.eid)
            self.assertNotEqual(patch.eid, patch2.eid)
            self.assertEqual(2, len(patch1.patch_revision))
            self.assertEqual(1, len(patch2.patch_revision))
            patch1_state = patch1.cw_adapt_to('IWorkflowable').state
            self.assertEqual('in-progress', patch1_state)
            patch1_wf_history = self.get_wf_history(patch1)
            self.assertEqual([], patch1_wf_history)
            patch2_state = patch1.cw_adapt_to('IWorkflowable').state
            self.assertEqual('in-progress', patch2_state)
            patch2_wf_history = self.get_wf_history(patch2)
            self.assertEqual([], patch2_wf_history)

    def test_fold(self):
        self.test_root_publishing()
        hg = self.hgrepo
        a = self.hg_append_to

        ### 7) fold
        #############################################################
        # create two draft changesets
        a('jungle', 'Cornelius')
        hg.commit('counselor')
        a('jungle', 'Pompadour')
        hg.commit('More counselor')
        self.refresh()
        # test we have two distinct patches
        with self.admin_access.client_cnx() as cnx:
            rev1 = self.all_revs(cnx)[-2]
            rev2 = self.all_revs(cnx)[-1]
            assert rev1.description == 'counselor'
            assert rev2.description == 'More counselor'
            self.assertTrue(rev1.reverse_patch_revision)
            self.assertTrue(rev2.reverse_patch_revision)
            patch1 = rev1.reverse_patch_revision[0].eid
            patch2 = rev2.reverse_patch_revision[0].eid
            self.assertNotEqual(patch1, patch2)

        # fold the two changesets
        hg.update('.~2')
        a('jungle', 'Cornelius')
        a('jungle', 'Pompadour')
        hg.commit('Counselor')
        pre1 = self.get_id('-3')
        pre2 = self.get_id('-2')
        succ = self.get_id('-1')
        self.hg('debugobsolete', pre1, succ)
        self.hg('debugobsolete', pre2, succ)
        self.refresh()
        # patch1 is folded (into patch2)
        with self.admin_access.client_cnx() as cnx:
            succ = self.all_revs(cnx)[-1]
            assert succ.description == 'Counselor'
            self.assertTrue(succ.reverse_patch_revision)
            patch = succ.reverse_patch_revision[0]
            morepatch = cnx.entity_from_eid(patch2)
            self.assertEqual(patch1, patch.eid)
            self.assertEqual(2, len(patch.patch_revision))
            self.assertEqual(1, len(morepatch.patch_revision))
            patch1_state = patch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('in-progress', patch1_state)
            wf1_history = self.get_wf_history(patch)
            self.assertEqual([], wf1_history)
            patch2_state = morepatch.cw_adapt_to('IWorkflowable').state
            self.assertEqual('folded', patch2_state)
            wf2_history = self.get_wf_history(morepatch)
            self.assertEqual(['folded'], wf2_history)

    def test_late_obsolete(self):
        hg = self.hgrepo
        a = self.hg_append_to

        # create a simple root changeset as draft
        a('jungle', 'Babar')
        hg.commit('Initial jungle content')
        self.refresh()

        # create an unrelated public cs
        hg.update('null')
        a('jungle', 'Babar')
        hg.commit('Amended jungle content')
        hg.phase('.', public=True)
        self.refresh()

        # add obsolete relation
        self.hg('debugobsolete', self.get_id('0'), self.get_id('1'))
        self.refresh()

        with self.admin_access.client_cnx() as cnx:
            rev0 = self.all_revs(cnx)[0]
            self.assertEqual(1, len(rev0.reverse_patch_revision))
            self.assertEqual(1, len(rev0.reverse_obsoletes))
            self.assertEqual(self.all_revs(cnx)[1].changeset, rev0.reverse_obsoletes[0].changeset)
            patch = rev0.reverse_patch_revision[0]
            self.assertEqual('applied', patch.cw_adapt_to('IWorkflowable').state)


class RepositoryNotificationTC(HgHelperTC):

    def test_repo_nosy_list(self):
        """Test that users in the repository nosylist get notified from new
        patches and their activity.
        """
        with self.admin_access.client_cnx() as cnx:
            self.create_user(cnx, 'Zephir', email=u'zephir@monke.ys',
                             interested_in=cnx.find('Repository').one())
            cnx.commit()
        hg = self.hgrepo
        self.hg_append_to('pub', 'whiskey')
        hg.commit('refresh', user='Cornelius', date='2012-01-01 00:00 +0300')
        self.refresh()
        with self.admin_access.client_cnx() as cnx:
            self.repoeid = cnx.find('Repository')[0][0]
            self.patcheid = cnx.find('Patch', patch_name='refresh')[0][0]
        self.set_description('new patch notification')
        self.assertEqual(len(MAILBOX), 1)
        yield self._check_new_patch_notification, MAILBOX[0]
        self.set_description('patch activity notification')
        MAILBOX[:] = []
        with self.admin_access.client_cnx() as cnx:
            patch = cnx.find('Patch', patch_name='refresh').one()
            patch.cw_adapt_to('IWorkflowable').fire_transition('accept', u'LGTM')
            cnx.commit()
        yield self.assertEqual, MAILBOX[0].recipients, ['zephir@monke.ys']

    def _check_new_patch_notification(self, mail):
        self.assertEqual(mail.recipients, ['zephir@monke.ys'])
        self.assertEqual(mail.subject,
                         '[mercurial repository #%s] new patch: refresh' % self.repoeid)
        self.assertMultiLineEqual(
            mail.message.get_payload(decode=True),
            '\n'.join([
                'url: http://testing.fr/cubicweb/patch/%s' % self.patcheid,
                '',
                '# HG changeset patch',
                '# User Cornelius',
                '# Date 1325365200 -10800',
                '#      Sun Jan 01 00:00:00 2012 +0300',
                '# Node ID bc2d1a04e9b830e21c16e0e9ad47e046bc912c29',
                '# Parent  0000000000000000000000000000000000000000',
                'refresh',
                '',
                'diff --git a/pub b/pub',
                'new file mode 100644',
                '--- /dev/null',
                '+++ b/pub',
                '@@ -0,0 +1,1 @@',
                '+whiskey',
                '',
                '',
            ])
        )


class PatchNotificationTC(HGRCMixin, CubicWebTC):

    @classmethod
    def setUpClass(cls):
        cls.__tmpdir = tempfile.mkdtemp(prefix='vcreview-test-')
        tarfile.open(cls.repo_path + '.tar').extractall(cls.__tmpdir)

    @classmethod
    def tearDownClass(cls):
        cls.cleanup()

    @classmethod
    def cleanup(cls):
        try:
            rmtree(cls.__tmpdir)
        except:
            pass

    @classproperty
    def full_repo_path(cls):
        return osp.join(cls.__tmpdir, u'repo')

    # XXX presetup
    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            # exist in the Mercurial repo
            self.create_user(cnx, 'Babar', ('users', 'reviewers', 'committers'),
                             email=u'babar@jungle.net')
            self.create_user(cnx, 'Victor', ('users', 'reviewers'),
                             email=u'victor@ratax.es')
            # unknown to the Mercurial repo
            self.create_user(cnx, 'Basile', ('users', 'reviewers'),
                             email=u'basile@ratax.es')
            cnx.commit()
            cnx.create_entity('Repository',
                              title=u'Jungle',
                              type=u'mercurial',
                              encoding=u'utf-8',
                              has_review=True,
                              local_cache=self.full_repo_path)
            cnx.commit()
        init_vcsrepo(self.repo)

    def assertEmailEqual(self, email, subject, recipient, content):
        self.assertEqual(email.subject, subject)
        self.assertEqual(email.recipients, [recipient])
        expected = content.splitlines()
        got = email.content.splitlines()
        # we should print a diff like -/+ stuff
        for idx, line in enumerate(expected):
            self.assertMatches('^' + line + '$', got[idx], idx)
        self.assertEqual(len(expected), len(got))

    def assertMatches(self, expected, got, lineno):
        """ expected is a regular expression """
        self.assertTrue(re.match(expected, got),
                'line %d: %r\n does not match:\n %r' % (lineno, got, expected))


    def test_nosy_list_author(self):
        """Test that author of a patch is automatically added to the nosylist"""
        with self.admin_access.client_cnx() as cnx:
            victor = cnx.execute('Any X WHERE X login "Victor"').get_entity(0, 0)
            all_nosy = set(e.eid for e in victor.reverse_nosy_list)
            rset = cnx.execute('Any P WHERE P patch_revision R, '
                               '            R author ILIKE "Victor%"')
            victor_patch = set(eid for eid, in rset)
            self.assertTrue(victor_patch.issubset(all_nosy))
            patch = rset.get_entity(0, 0)
            self.assertEqual(victor.eid, patch.creator.eid)
            self.assertEqual(victor.name(), patch.dc_authors())

    def test_nosy_list_task_author_commenter(self):
        """Test that users adding tasks or comments are added in nosylist"""
        with self.admin_access.repo_cnx() as cnx:
            cornelius_eid = self.create_user(cnx, 'Cornelius').eid
            zephir_eid = self.create_user(cnx, 'Zephir').eid
            cnx.commit()
        with self.new_access('Cornelius').client_cnx() as cnx:
            rset = cnx.find('Revision', description='Add more elephants children')
            rev = rset.get_entity(1, 0)
            patch = rev.reverse_patch_revision[0]
            patch_eid = patch.eid
            task = cnx.create_entity('Task', title=u'todo', reverse_has_activity=patch)
            task_eid = task.eid
            point = cnx.create_entity('InsertionPoint', lid=0, point_of=rev)
            cnx.commit()
        with self.new_access('Zephir').client_cnx() as cnx:
            comment_eid = cnx.create_entity(
                'Comment', comments=task_eid, content=u'coucou').eid
            cnx.commit()
        with self.new_access('Babar').client_cnx() as cnx:
            cnx.create_entity(
                'Comment', comments=comment_eid, content=u'salut')
            cnx.commit()
        with self.admin_access.repo_cnx() as cnx:
            patch = cnx.find('Patch', eid=patch_eid).one()
            patch_nosylist = set(x.login for x in patch.nosy_list)
        self.set_description('author of a task in patch nosylist')
        yield self.assertIn, 'Cornelius', patch_nosylist
        self.set_description('commenter of a task in patch nosylist')
        yield self.assertIn, 'Zephir', patch_nosylist
        self.set_description('commenter of a task comment in patch nosylist')
        yield self.assertIn, 'Babar', patch_nosylist

    def test_task_notification(self):
        with self.new_access('Victor').client_cnx() as cnx:
            rql = 'Any R ORDERBY R DESC WHERE R is Revision'
            rev = cnx.execute(rql).get_entity(1, 0)
            assert rev.description == 'Add more elephants children'
            patch = rev.reverse_patch_revision[0]
            assert patch.patch_name == 'Add more elephants children'
            task = cnx.create_entity('Task', title=u'todo', reverse_has_activity=patch)
            cnx.commit()
            assert len(MAILBOX) == 1, MAILBOX
            self.assertEmailEqual(
                MAILBOX[0],
                '[Jungle] patch task: Add more elephants children - todo',
                'victor@ratax.es',
                u'\n'.join([u'', u'todo\n\n\n',
                            u'url: {}'.format(patch.absolute_url())]))
            MAILBOX[:] = []
            point = cnx.create_entity('InsertionPoint', lid=123, point_of=rev)
            task = cnx.create_entity('Task', title=u'needs work!', reverse_has_activity=point)
            cnx.commit()
            self.assertEmailEqual(
                MAILBOX[0],
                '[Jungle] patch task: Add more elephants children - needs work!',
                'victor@ratax.es',
                u'\n'.join([u'', u'needs work!\n\n\n',
                            u'url: {}#navcontentbottom{}'.format(
                                rev.absolute_url(), point.eid)]))

    def test_reviewer_email(self):
        with self.admin_access.client_cnx() as cnx:
            rql = 'Any R ORDERBY R DESC WHERE R is Revision'
            rev = cnx.execute(rql).get_entity(1, 0)
            assert rev.description == 'Add more elephants children'
            patch = rev.reverse_patch_revision[0]
            assert patch.patch_name == 'Add more elephants children'
            patch.cw_adapt_to('IWorkflowable').fire_transition('ask review')
            cnx.commit()

            self.assertEqual(1, len(patch.patch_reviewer))
            reviewer = patch.patch_reviewer[0]
            self.assertIn(reviewer.login, ('Babar', 'Basile'))
            nosyed = set(u.login for u in patch.nosy_list)
            self.assertEqual(2, len(nosyed))
            self.assertIn(reviewer.login, nosyed)

            MAILBOX[:] = []
            patch.cw_adapt_to('IWorkflowable').fire_transition('ask rework',
                                                               u'coin coin')
            cnx.commit()
            self.assertEqual(2, len(MAILBOX))

            MAILBOX[:] = []
            # patch are only send on ask review, so we reask for review
            patch.cw_adapt_to('IWorkflowable').fire_transition('ask review',
                                                               u'coin coin')
            cnx.commit()
            self.assertEqual(2, len(MAILBOX))
            MAILBOX.sort(key=lambda e: 'victor@ratax.es' in e.recipients)
            review_email = MAILBOX[0]
            self.assertEmailEqual(review_email,
                                  '[Jungle] patch pending-review: Add more elephants children',
                                  [e.address for e in reviewer.primary_email][0],
                                  '\n'
                                  'admin changed status from <in-progress> to <pending-review> for entity\n'
                                  "'Add more elephants children'\n"
                                  '\n'
                                  'coin coin\n'
                                  '\n'
                                  '\(No recorded task\)\n'
                                  '\n'
                                  'url: http://testing.fr/cubicweb/patch/\d+\n'
                                  '\n'
                                  'reviewer: Ba(bar|sile)\n'
                                  '\n'
                                  '# HG changeset patch\n'
                                  '# User Victor <victor@ratax.es>\n'
                                  '# Date 1361805588 0\n'
                                  '%s' # optional human-readable date
                                  '# Node ID bfa5da61c82507b76dae7927b4fce2aa4a0c9c6f\n'
                                  '# Parent  549cd3165597cdd5c9d8fa26e83e63483fcfb4b5\n'
                                  'Add more elephants children\n'
                                  '\n'
                                  'diff --git a/babar.py b/babar.py\n'
                                  '--- a/babar.py\n'
                                  '\+\+\+ b/babar.py\n'
                                  '@@ -5,10 \+5,15 @@\n'
                                  " elephants = \['Babar',\n"
                                  "              'Celeste',\n"
                                  "              'Cornelius',\n"
                                  "              'Pompadour',\n"
                                  "              'Poutifour',\n"
                                  "\+             'Arthur'\n"
                                  "\+             'Pom',\n"
                                  "\+             'Flore',\n"
                                  "\+             'Alexandre',\n"
                                  "\+             'Isabelle',\n"
                                  '              \]\n'
                                  ' \n'
                                  ' \n'
                                  " if __name__ == '__main__':\n"
                                  "     print 'Those are elephants:'\n"
                                  '\n' %
                                  '#      Mon Feb 25 15:19:48 2013 \+0000\n')
            # add task
            MAILBOX[:] = ()
            task = cnx.create_entity('Task', title=u'todo', reverse_has_activity=patch)
            point = cnx.create_entity('InsertionPoint', lid=0, point_of=rev)
            cnx.create_entity('Task', title=u'my point', reverse_has_activity=point)
            cnx.commit()
            self.assertEqual(4, len(MAILBOX))
            # reask review
            MAILBOX[:] = ()
            patch.cw_adapt_to('IWorkflowable').fire_transition('ask rework',
                                                               u'tasked ø')
            cnx.commit()
            self.assertEqual(2, len(MAILBOX))
            MAILBOX.sort(key=lambda e: 'victor@ratax.es' in e.recipients)
            review_email = MAILBOX[0]
            self.assertEmailEqual(review_email,
                                  '[Jungle] patch in-progress: Add more elephants children',
                                  [e.address for e in reviewer.primary_email][0],
                                  '\n'
                                  'admin changed status from <pending-review> to <in-progress> for entity\n'
                                  "'Add more elephants children'\n"
                                  '\n'
                                  u'tasked ø\n'
                                  '\n'
                                  'remaining tasks:\n'
                                  '\n'
                                  '- my point\n'
                                  '- todo\n'
                                  '\n'
                                  'url: http://testing.fr/cubicweb/patch/\d+\n'
                                  '\n'
                                  'reviewer: Ba(bar|sile)\n'
                                  '\n'
                                  '\n')

            # Test that comments are emailed too
            MAILBOX[:] = ()
            point.cw_clear_all_caches()
            comment = cnx.create_entity('Comment', comments=task,
                                            content=u'Babar comments')
            cnx.commit()
            MAILBOX.sort(key=lambda e: 'victor@ratax.es' in e.recipients)
            self.assertEqual(2, len(MAILBOX))
            review_email = MAILBOX[0]
            self.assertEmailEqual(review_email,
                                  'new comment for Task todo',
                                  [e.address for e in reviewer.primary_email][0],
                                  'Babar comments\n'
                                  '\n\n'
                                  'i18n_by_author_field: admin\n'
                                  'url: http://testing.fr/cubicweb/task/%i' % task.eid)

    def test_nonascii(self):
        with self.admin_access.client_cnx() as cnx:
            self.assertEqual(0, len(MAILBOX))
            rql = 'Any R ORDERBY R DESC WHERE R is Revision'
            rev = cnx.execute(rql).get_entity(0, 0)
            patch = rev.reverse_patch_revision[0]
            self.assertEqual(0, len(MAILBOX))
            patch_flow = patch.cw_adapt_to('IWorkflowable')
            patch_flow.fire_transition('ask review', u'beep beep')
            cnx.commit()
            # 3: 2x "patch pending-review" + 1x "patch assigned"
            self.assertEqual(3, len(MAILBOX))
            patch_flow.fire_transition('ask rework', u'coin coin')
            cnx.commit()
            self.assertEqual(5, len(MAILBOX))
            patch_flow.fire_transition('ask review', u'beep beep')
            cnx.commit()
            self.assertEqual(7, len(MAILBOX))
            patch_flow.fire_transition('ask rework', u'coin coin')
            cnx.commit()
            self.assertEqual(9, len(MAILBOX))

            MAILBOX[:] = []
            # emails are only sent on ask review, so we ask again for review
            patch.cw_adapt_to('IWorkflowable').fire_transition('ask review',
                                                               u'coin coin')
            cnx.commit()
            self.assertEqual(2, len(MAILBOX))

PatchNotificationTC.repo_path = osp.join(PatchNotificationTC.datadir,
                                         u'demo-repo-final')


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2011-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-vcreview schema"""

__docformat__ = "restructuredtext en"
_ = unicode

from yams.buildobjs import (EntityType,
                            RelationDefinition,
                            String,
                            Int,
                            Boolean,
                            SubjectRelation)

from cubicweb.schema import (WorkflowableEntityType,
                             RQLConstraint,
                             RQLVocabularyConstraint,
                             RRQLExpression)

from cubes.vcsfile.schema import Repository


Repository.add_relation(Boolean(default=False, required=True),
                        name='has_review')

# the patch entity #############################################################


_SAME_REPO_CONSTRAINT = '''NOT EXISTS(
    S patch_revision RE,
    RE from_repository RO,
    O from_repository R,
    NOT R identity RO
)'''

class Patch(WorkflowableEntityType):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': (), # created by an internal session in a looping task
        'update': ('managers', 'users'),
        'delete': ('managers',),
        }
    patch_name = String(
        required=True,
        description=_('name of the patch in the repository'),
        fulltextindexed=True)
    patch_revision = SubjectRelation('Revision',
                                     __permissions__={
                                         'read': ('managers', 'users', 'guests'),
                                         'add': (), # created by an internal session in a looping task
                                         'delete': (),
                                         },
                                     constraints=[RQLConstraint(_SAME_REPO_CONSTRAINT)],
                                     cardinality='+?')


class repository_committer(RelationDefinition):
    """integrator (review level 2)"""
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers',),
        'delete': ('managers',),
        }
    subject = 'Repository'
    object = 'CWUser'

class repository_reviewer(RelationDefinition):
    """(review level 1)"""
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', 'committers',
                RRQLExpression('S repository_committer U')),
        'delete': ('managers', 'committers',
                   RRQLExpression('S repository_committer U')),
        }
    subject = 'Repository'
    object = 'CWUser'

class patch_reviewer(RelationDefinition):
    """The good fellow reviewing a single patch"""
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', 'committers',
                RRQLExpression('S patch_revision RE, RE from_repository R, R repository_committer U')),
        'delete': ('managers', 'committers',
                   RRQLExpression('S patch_revision RE, RE from_repository R, R repository_committer U')),
        }
    subject = 'Patch'
    object = 'CWUser'
    constraints = [RQLVocabularyConstraint('EXISTS(O in_group G, G name "reviewers") '
                                 'OR EXISTS(S patch_revision RE, '
                                 '          RE from_repository R, '
                                 '          R repository_reviewer O)')]

class patch_committer(RelationDefinition):
    """The level 2 guy in charge of getting the patch published"""
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', 'committers',
                RRQLExpression('S patch_revision RE, RE from_repository R, R repository_committer U')),
        'delete': ('managers', 'committers',
                   RRQLExpression('S patch_revision RE, RE from_repository R, R repository_committer U')),
        }
    subject = 'Patch'
    object = 'CWUser'
    constraints = [RQLVocabularyConstraint('EXISTS(O in_group G, G name "committers") '
                                 'OR EXISTS(S patch_revision RE, RE from_repository R, R repository_committer O)')]


class use_global_groups(RelationDefinition):
    """Use the ``reviewer`` global group to pick reviewer

    When false, only explicit reviewer set with ``repository_reviewer`` are
    taken in account.
    """
    subject = 'Repository'
    object = 'Boolean'
    cardinality = '11'
    default = True


# ability to reference diff chunks #############################################

class InsertionPoint(EntityType):
    """diff hunk"""
    __unique_together__ = [('lid', 'point_of')]
    lid = Int(indexed=True, required=True)

    point_of = SubjectRelation('Revision', cardinality = '1*',
                               composite = 'object', inlined = True)


# activities ###################################################################

class has_activity(RelationDefinition):
    subject = ('Patch', 'InsertionPoint')
    object = 'Task'
    cardinality = '*?'
    composite = 'subject'


# comments #####################################################################

class comments(RelationDefinition):
    subject = 'Comment'
    object = ('Task',)

# notification #################################################################

# XXX Comment should be in list, but this will conflict with the forge cube when
# using vcreview
class nosy_list(RelationDefinition):
    subject = ('Repository', 'Patch', 'Revision', 'Task', 'InsertionPoint')
    object = 'CWUser'

class interested_in(RelationDefinition):
    subject = 'CWUser'
    object = ('Repository', 'Patch')



# -*- coding: utf-8 -*-

# copyright 2011-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import os
from os import path as osp
from shutil import rmtree
import re
import tempfile
from subprocess import check_call, Popen, PIPE
import hglib

from cubicweb.devtools.testlib import CubicWebTC
from cubes.vcsfile import bridge
from cubes.vcsfile.testutils import HGRCMixin

os.environ['HGRCPATH'] = os.devnull


class HgHelperTC(HGRCMixin, CubicWebTC):
    hgconfig = HGRCMixin.hgconfig.copy()
    hgconfig['extensions']['mq'] = ''
    verbose_hg = False

    def setup_database(self):
        self.tmprepo = tempfile.mkdtemp()
        hglib.init(self.tmprepo)
        self.hgrepo = hglib.open(self.tmprepo)
        with self.admin_access.client_cnx() as cnx:
            self.cwhgrepo = cnx.create_entity('Repository',
                                              type=u'mercurial',
                                              encoding=u'utf-8',
                                              import_revision_content=True,
                                              local_cache=unicode(self.tmprepo),
                                              has_review=True).eid
            cnx.commit()

    def tearDown(self):
        self.hgrepo.close()
        rmtree(self.tmprepo, ignore_errors=True)
        return super(HgHelperTC, self).tearDown()

    def hg(self, *args):
        return self.hgrepo.rawcommand(args)

    def hg_append_to(self, path, content='', add=True):
        """Add a new line to a file in the repo"""
        path = osp.join(self.tmprepo, path)
        with open(path, 'a') as f:
            f.write(content)
            f.write('\n')
        if add:
            self.hgrepo.add(path)

    def refresh(self):
        """Have cubicweb read the on disk repo"""
        with self.repo.internal_cnx() as cnx:
            bridge.import_content(cnx, raise_on_error=True)
            cnx.commit()

    def all_revs(self, cnx):
        """Return all revisions that exist in the test repo entity"""
        rql = 'Revision R ORDERBY R WHERE R from_repository RE, RE eid %(re)s'
        rset = cnx.execute(rql, {'re': self.cwhgrepo})
        return list(rset.entities())

    def get_id(self, rev):
        """return the full hex of the full id of a changeset"""
        cmd = ['id', '--debug', '--id', '--rev', rev]
        if self.verbose_hg:
            print ' '.join(cmd)
        rid = self.hgrepo.rawcommand(cmd).strip()
        assert len(rid) == 40, rid
        return rid

    @staticmethod
    def get_wf_history(patch):
        return [tr.new_state.name
                for tr in patch.cw_adapt_to('IWorkflowable').workflow_history]



# copyright 2011-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-vcreview entity's classes"""

from datetime import datetime

from logilab.common.decorators import monkeypatch

from cubicweb.entities import AnyEntity
from cubicweb.view import EntityAdapter
from cubicweb.predicates import is_instance

from cubes.comment import entities as comment
from cubes.task import entities as task

from cubes.vcreview import users_by_author

class InsertionPoint(AnyEntity):
    __regid__ = 'InsertionPoint'
    @property
    def parent(self):
        return self.point_of[0]

_TIP_MOST_REVS = ('Any TIP, H '
                  'ORDERBY H ASC, TIP DESC'
                  'WHERE P eid %(p)s, '
                  '      P patch_revision TIP, '
                  '      TIP hidden H, '
                  '      NOT EXISTS(R obsoletes TIP, '
                  '                 P patch_revision R)')

class Patch(AnyEntity):
    __regid__ = 'Patch'

    def dc_title(self):
        return self.patch_name

    @property
    def repository(self):
        return self.patch_revision[0].from_repository[0]

    @property
    def revisions(self):
        return sorted(self.patch_revision, key=lambda x: x.revision)

    def patch_files(self):
        return set(vc.file.path for vc in self.patch_revision)

    def tip(self):
        return self._cw.entity_from_eid(self.all_tips()[0][0])

    def all_tips(self):
        return list(self._cw.execute(_TIP_MOST_REVS, {'p': self.eid}))

    final_states = set( ('folded', 'applied', 'rejected') )

    @property
    def is_final(self):
        return self.cw_adapt_to('IWorkflowable').state in self.final_states

    def email_export(self, size=None):
        """Return an export of patch possibly truncated at `size` for email
        embedding.
        """
        patch = self.tip().export().decode(self.repository.encoding, 'replace')
        if size is not None and len(patch) > size:
            patch = patch[:size] + '\n@@ truncated patch @@'
        return patch


@monkeypatch(task.Task, 'activity_of')
@property
def activity_of(self):
    return self.reverse_has_activity and self.reverse_has_activity[0]


@monkeypatch(task.Task, 'patch')
@property
def patch(self):
    """The patch associated with this Task"""
    concerned = self.activity_of
    if not concerned:
        return None
    if concerned.cw_etype == 'Patch':
        return concerned
    else:
        return concerned.point_of[0]


class CommentITreeAdapter(comment.CommentITreeAdapter):
    def path(self):
        path = super(CommentITreeAdapter, self).path()
        rootrset = self._cw.eid_rset(path[0])
        if rootrset.description[0][0] == 'InsertionPoint':
            path.insert(0, rootrset.get_entity(0, 0).parent.eid)
        return path


class PatchReviewBehaviour(EntityAdapter):
    __regid__ = 'IPatchReviewControl'
    __select__ = is_instance('Patch')

    def reviewers_rset(self):
        # ensure patch author can't review his own patch
        return self._cw.execute(
            'Any U, MAX(D) GROUPBY U '
            'WHERE P eid %(p)s, '
            '      P patch_revision RE, '
            '      RE from_repository R, '
            '      NOT P created_by U, '
            '      U in_state S, NOT S name "deactivated", '
            '      X? patch_reviewer U, '
            '      X creation_date D, '
            '      EXISTS(R use_global_groups TRUE, U in_group G, G name "reviewers") '
            '      OR EXISTS(R repository_reviewer U)',
            {'p': self.entity.eid})

    def set_reviewers(self):
        if self.entity.patch_reviewer:
            return
        # look for patch authors, so we don't ask them to review their own patch
        authors = set()
        for rev in self.entity.patch_revision:
            authors.update(users_by_author(self._cw.execute, rev.author))

        for ueid,_ in sorted(self.reviewers_rset(), key=lambda row: row[1] or datetime.min):
            if ueid in authors:
                continue
            self._cw.execute('SET P patch_reviewer U WHERE P eid %(p)s, U eid %(u)s',
                             {'p': self.entity.eid, 'u': ueid})
            break


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, (CommentITreeAdapter,))
    vreg.register_and_replace(CommentITreeAdapter, comment.CommentITreeAdapter)



create_entity('CWGroup', name=u'reviewers')
create_entity('CWGroup', name=u'committers')
add_relation_type('repository_committer')
add_relation_type('repository_reviewer')
add_relation_type('patch_reviewer')
add_relation_type('patch_depends_on')


pwf = get_workflow_for('Patch', ask_confirm=False)
inprogress = pwf.state_by_name('in-progress')
pending    = pwf.state_by_name('pending-review')
applied    = pwf.state_by_name('applied')
rejected   = pwf.state_by_name('rejected')
folded     = pwf.state_by_name('folded')
deleted    = pwf.state_by_name('deleted')

reviewed   = pwf.add_state('reviewed')
pwf.add_transition(_('apply'), (inprogress, pending, deleted), applied,
                   ('managers',))

accept = pwf.transition_by_name('accept')
accept.set_relations(destination_state=reviewed)
reviewed.set_relations(allowed_transition=[pwf.transition_by_name('refuse'),
                                           pwf.transition_by_name('fold'),
                                           pwf.transition_by_name('apply'),
                                           pwf.transition_by_name('reject')])
for trname in ('accept', 'refuse', 'fold'):
    pwf.transition_by_name(trname).set_permissions(
        ('reviewers',),
        'X patch_repository R, R repository_reviewer U',
        reset=False)

for transition in pwf.reverse_transition_of:
    transition.set_permissions(
        ('committers',), 'X patch_repository R, R repository_committer U',
        reset=False)



pwf = get_workflow_for('Patch', ask_confirm=False)
outdated = pwf.state_by_name('outdated')
reviewed = pwf.state_by_name('reviewed')
pwf.add_transition(_('reopen/accept'), outdated, reviewed, ('managers', ))



conditions = [
    # global reviewer / committer
    'Z has_activity X, U in_group G, G name IN ("reviewers", "committers")',
    # repository reviewer / committer
    'P has_activity X, P patch_repository R, EXISTS(R repository_reviewer U) OR EXISTS(R repository_committer U)',
    'IP has_activity X, IP point_of VC, P patch_revision VC, P patch_repository R, EXISTS(R repository_reviewer U) OR EXISTS(R repository_committer U)',
    # patch owner
    'P has_activity X, P owned_by U',
    'IP has_activity X, IP point_of VC, P patch_revision VC, P owned_by U',
    ]
task_wf = get_workflow_for('Task')
task_wf.transition_by_name('start').set_permissions(conditions=conditions, reset=False)
task_wf.transition_by_name('done').set_permissions(conditions=conditions, reset=False)



# copyright 2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-vcreview components (boxes, pluggable sections)"""

__docformat__ = "restructuredtext en"
_ = unicode

from logilab.mtconverter import xml_escape

from cubicweb.predicates import (is_instance, relation_possible,
                                score_entity, match_form_params)
from cubicweb import tags, uilib, view
from cubicweb.web import component, form, formwidgets as wdg
from cubicweb.view import EntityView
from cubicweb.web.views import (workflow, idownloadable, ajaxedit,
                                forms, formrenderers, uicfg)
from cubicweb.web.views.ajaxcontroller import ajaxfunc

from cubes.comment import views as comment
from cubes.vcreview.site_cubicweb import COMPONENT_CONTEXT

_afs = uicfg.autoform_section
_affk = uicfg.autoform_field_kwargs
_abaa = uicfg.actionbox_appearsin_addmenu
_pvs = uicfg.primaryview_section



# activities ###################################################################

_pvs.tag_subject_of(('*', 'has_activity', '*'), 'hidden') # component
_affk.tag_object_of(('*', 'has_activity', 'Task'),
                    {'widget': wdg.HiddenInput})


class AddActivityFormView(form.FormViewMixIn, view.EntityView):
    __regid__ = 'vcreview.addactivityform'
    __select__ = (relation_possible('has_activity', 'subject', 'Task', 'add')
                  | match_form_params('etype'))

    def call(self, **kwargs):
        if self.cw_rset is None:
            entity = self._cw.vreg['etypes'].etype_class(self._cw.form['etype'])(self._cw)
            entity.eid = self._cw.form['tempEid']
            self.entity_call(entity)
        else:
            super(AddActivityFormView, self).call(**kwargs)

    def entity_call(self, container):
        self._cw.add_js(('cubicweb.edition.js', 'cubes.vcreview.js'))
        activity = self._cw.vreg['etypes'].etype_class('Task')(self._cw)
        activity.eid = self._cw.varmaker.next()
        okjs = uilib.js.addActivity(container.eid, not container.has_eid(),
                                    self.cw_extra_kwargs.get('context'))
        canceljs = "jQuery('#activity%sSlot').remove()" % container.eid
        form, formvalues = ajaxedit.ajax_composite_form(
            container, activity, 'has_activity', okjs, canceljs,
            dict(display_fields=(('title', 'subject'),
                                 ('description', 'subject'),
                                 ('has_activity', 'object'))))
        self.w(u'<div id="activity%sSlot">' % container.eid)
        form.render(w=self.w, main_form_title=u'', formvalues=formvalues)
        self.w(u'</div>')


class ActivityComponent(component.EntityCtxComponent):
    """a component to display an html section displaying activities related to
    an object
    """
    __regid__ = 'vcreview.activitysection'
    __select__ = (component.EntityCtxComponent.__select__
                  & relation_possible('has_activity', 'subject', 'Task'))

    context = 'navcontentbottom'

    def render_body(self, w):
        req = self._cw
        req.add_js('cubicweb.ajax.js')
        if self.entity.has_eid():
            self.display_existing_tasks(w)
            eid = self.entity.eid
        else:
            eid = None
        rdef = self.entity.e_schema.rdef('has_activity', 'subject', 'Task')
        if (req.vreg.schema['Task'].has_perm(req, 'add') and
            rdef.has_perm(req, 'add', fromeid=eid)):
            url = self.lazy_view_holder(w, self.entity, 'vcreview.addactivityform')
            w(u'(<a href="%s">%s</a>)' % (xml_escape(url), req._('add a task')))

    existing_tasks_rql = (
        u'Any T,TD,T,TCD,U, TS,TT,TDF,UL,UF,US ORDERBY TCD DESC WHERE '
        'X has_activity T, T description TD, T in_state TS,'
        'T creation_date TCD, T created_by U?,'
        'T title TT, T description_format TDF, '
        'U login UL, U firstname UF, U surname US, X eid %(x)s')

    def display_existing_tasks(self, w):
        rset = self._cw.execute(self.existing_tasks_rql, {'x': self.entity.eid})
        if rset:
            w(u'<h4>%s</h4>' % self._cw._('Activities'))
            self._cw.view('vcreview.task-view', rset, w=w)


class TaskView(EntityView):
    """a component to display an html section displaying activities related to
    an object
    """
    __regid__ = 'vcreview.task-view'
    __select__ = EntityView.__select__ & is_instance('Task')

    def cell_call(self, row, col):
        self._cw.add_css('cubes.vcreview.css')
        self._cw.add_js('cubes.vcreview.js')
        assert col == 0
        task = self.cw_rset.get_entity(row, col)
        iwf  = task.cw_adapt_to('IWorkflowable')
        assert iwf.state in ('todo', 'done'), iwf.state
        self.w(u'<div class="vcreview_task task_%s" data-eid="%s">' % (iwf.state, task.eid))
        self.w(u'<h5>')
        self.w(u'<span class="vcreview_taskstate">')
        self.w(iwf.printable_state)
        self.w(u'</span>')
        self.w(u'<span class="vcreview_tasktitle">')
        self.wview(u'vcreview.task-link', rset=self.cw_rset, row=row, col=col)
        self.w(u'</span>')
        if self._can_mark_as_done(iwf):
            self.w(tags.input(type='button', value='mark as done', klass="task_is_done"))
        self.w(u'</h5>')
        self.w(u'<p class="vcreview_taskinfo">Created ')
        self.w(task.printable_value('creation_date'))
        if task.created_by:
            self.w(u' by ')
            self.wview(u'incontext', rset=task.created_by[0].as_rset())
        self.w(u'</p>')
        self.w(u'<div id="vcreview_taskdescription">')
        self.w(task.view('reledit', rtype='description'))
        self.w(u'</div>')

        self.w(u'<div id="navcontentbottom%s">' % task.eid)
        comp = self._cw.vreg['ctxcomponents'].select_or_none('commentsection',
            self._cw,
            rset=task.as_rset(),
            entity=task,
            context='navcontentbottom',
            __cache=None)
        if comp is not None:
            comp.render(w=self.w)
        self.w(u'</div>')

        self.w(u'</div>')

    @staticmethod
    def _can_mark_as_done(iwf):
        if iwf.state != 'todo':
            return False
        for transition in iwf.possible_transitions():
            if transition.name == 'done':
                return True
        else:
            return False # task workflow changed or no permission to go through the transition

@ajaxfunc
def vcreview_change_state(self, eid):
    entity = self._cw.entity_from_eid(eid)
    entity.cw_adapt_to('IWorkflowable').fire_transition('done')


class PatchActivityComponent(ActivityComponent):
    __select__ = ActivityComponent.__select__ & is_instance('Patch')
    existing_tasks_rql = (
        u'Any T,TD,T,TCD,U, TS,TT,TDF,UL,UF,US ORDERBY TCD DESC '
        'WITH T,TD,TS,TCD,U,TT,TDF,UL,UF,US BEING ('
        '(Any T,TD,TS,TCD,U,TT,TDF,UL,UF,US WHERE '
        'X has_activity T, T description TD, T in_state TS,'
        'T creation_date TCD, T created_by U?, '
        'T title TT, T description_format TDF, '
        'U login UL, U firstname UF, U surname US, X eid %(x)s)'
        ' UNION '
        '(Any T,TD,TS,TCD,U,TT,TDF,UL,UF,US WHERE '
        'X patch_revision RE, IP point_of RE, IP has_activity T, '
        'T description TD, T in_state TS, T creation_date TCD, '
        'T created_by U?, T title TT, T description_format TDF, '
        'U login UL, U firstname UF, U surname US, X eid %(x)s)'
        ')')


workflow.WFHistoryVComponent.order = 10
ActivityComponent.order = 20
comment.CommentSectionVComponent.order = 30




class TaskLinkView(form.FormViewMixIn, view.EntityView):
    __regid__ = 'vcreview.task-link'
    __select__ = is_instance('Task')
    def entity_call(self, entity):
        self.w(entity.view('incontext'))
        if entity.reverse_has_activity[0].__regid__ == 'InsertionPoint' \
               and self._cw.form.get('vid') == 'vcreview.patch.tab_main':
            ip = entity.reverse_has_activity[0]
            purl = ip.parent.absolute_url()
            url = u'%s#%s%s' % (purl, COMPONENT_CONTEXT, ip.eid)
            self.w(u' (<a href="%s">%s</a>)' % (url, self._cw._('see in context')))




add_attribute('Repository', 'use_global_groups')
sync_schema_props_perms('patch_name')
add_relation_definition('Comment', 'comments', 'Task')

rql('''
INSERT Task T:
    T title "placeholder Task",
    T description "house free comment after 1.1.0 migration",
    P has_activity T,
    C comments T
WHERE
    C comments P,
    P is Patch''')
drop_relation_definition('Comment', 'comments', 'Patch')

rql('''
INSERT Task T:
    T title "placeholder Task",
    T description "house free comment after 1.1.0 migration",
    IP has_activity T,
    C comments T
WHERE
    C comments IP,
    IP is InsertionPoint''')
drop_relation_definition('Comment', 'comments', 'InsertionPoint')
commit()



# copyright 2011-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import os.path as osp
from itertools import count, izip
from subprocess import CalledProcessError

from logilab.common.testlib import TestCase, unittest_main
import cubicweb.devtools
from cubes.vcreview import compare_patches

from unittest_hooks import HgHelperTC

initialpatch="""# HG changeset patch
# User Victor <victor@ratax.es>
# Date 1362585681 -3600
# Node ID 70c874b4e12b01fc2a51c2dfdebc5c61b5da467c
# Parent  fbb870448944fadbdb30104109541a3316438717
[pkg] prepare release 1.0.0

pkginfo was already changed.

diff --git a/debian/changelog b/debian/changelog
--- a/debian/changelog
+++ b/debian/changelog
@@ -1,3 +1,10 @@
+cubicweb-vcreview (1.0.0-1) unstable; urgency=low
+
+  * new upstream release
+  * drop MQ backend in favor of changeset evolution
+
+ -- Pierre-Yves David <pierre-yves.david@logilab.fr>  Wed, 06 Mar 2013 16:59:11 +0100
+
 cubicweb-vcreview (0.10.1-1) unstable; urgency=low
 
   * new upstream release
diff --git a/debian/control b/debian/control
--- a/debian/control
+++ b/debian/control
@@ -2,14 +2,14 @@ Source: cubicweb-vcreview
 Section: web
 Priority: optional
 Maintainer: LOGILAB S.A. (Paris, FRANCE) <contact@logilab.fr>
-Build-Depends: debhelper (>= 5.0.37.1), python (>=2.4), python-dev (>=2.4)
+Build-Depends: debhelper (>= 5.0.37.1), python (>=2.6)
 Standards-Version: 3.8.0
 
 
 Package: cubicweb-vcreview
 Architecture: all
-Depends: cubicweb-common (>= 3.14.0),
-  cubicweb-vcsfile (>= 1.7.1), cubicweb-comment (>= 1.8.0),
+Depends: cubicweb-common (>= 3.15.8),
+  cubicweb-vcsfile (>= 1.12.0), cubicweb-comment (>= 1.8.0),
   cubicweb-task, cubicweb-nosylist
 Breaks: cubicweb-trackervcs (<< 1)
 Description: patch review system on top of vcsfile
"""

otherpatch="""# HG changeset patch
# User Victor <victor@ratax.es>
# Date 1362588960 -3600
# Node ID 3e6dbe86678c30c2a77a01da65f2842de287cb54
# Parent  46b2b7ffaf2096953433ba393c20c645eb11ddcd
Added tag cubicweb-vcreview-version-1.0.0 for changeset 46b2b7ffaf20

diff --git a/.hgtags b/.hgtags
--- a/.hgtags
+++ b/.hgtags
@@ -40,3 +40,4 @@ 2066c270a15bf59b7b6294353bb4e7c47e12df77
 3bd0fd59996885104d7a9435a8f665e79f4bac88 cubicweb-vcreview-debian-version-0.10.0-1
 f49629dd7ea14a93269b35d94b600401a885b30f cubicweb-vcreview-version-0.10.1
 e7005fcce2cdebc4875d147daad2d121f5e86146 cubicweb-vcreview-debian-version-0.10.1-1
+46b2b7ffaf2096953433ba393c20c645eb11ddcd cubicweb-vcreview-version-1.0.0
"""


class ComparePatch(TestCase):
    def assertSamePatch(self, old, new):
        self.assertTrue(compare_patches(old, new))

    def assertDiffPatch(self, old, new):
        self.assertFalse(compare_patches(old, new))

    def test_identity(self):
        self.assertSamePatch(initialpatch, initialpatch)

    def test_unrelated(self):
        self.assertDiffPatch(initialpatch, otherpatch)

    def test_user_change(self):
        victor = '# User Victor <victor@ratax.es>'
        assert victor in initialpatch
        podular = '# User Podular <podular@elephan.tz>'
        changed = initialpatch.replace(victor, podular)
        self.assertDiffPatch(initialpatch, changed)

    def test_node_change(self):
        ihash = '70c874b4e12b01fc2a51c2dfdebc5c61b5da467c'
        assert ihash in initialpatch
        nhash = 'kjcndsalkjncds'
        changed = initialpatch.replace(ihash, nhash)
        self.assertSamePatch(initialpatch, changed)

    def test_desc_change(self):
        idesc = 'pkginfo was already changed.'
        assert idesc in initialpatch
        ndesc = 'nothing'
        changed = initialpatch.replace(idesc, ndesc)
        self.assertDiffPatch(initialpatch, changed)

    def test_add_hunk(self):
        changed = initialpatch + otherpatch.split('\n\n')[1]
        self.assertDiffPatch(initialpatch, changed)

    def test_filename_change(self):
        ifn = 'debian/changelog'
        assert ifn in initialpatch
        changed = initialpatch.replace(ifn, 'debian/rules')
        self.assertDiffPatch(initialpatch, changed)

    def test_context_change(self):
        iprio = 'Priority: optional'
        assert iprio in initialpatch
        changed = initialpatch.replace(iprio, 'Priority: useless')
        self.assertDiffPatch(initialpatch, changed)

    def test_minus_change(self):
        iminus = '-Depends: cubicweb-common (>= 3.14.0),'
        assert iminus in initialpatch
        changed = initialpatch.replace(iminus,
                                       '-Depends: cubicweb-common (>= 3.14.1),')
        self.assertDiffPatch(initialpatch, changed)

    def test_plus_change(self):
        iplus = '+  * new upstream release'
        assert iplus in initialpatch
        changed = initialpatch.replace(iplus, '+  * bing')
        self.assertDiffPatch(initialpatch, changed)

class DetectRebaseTC(HgHelperTC):

    verbose_hg = False

    def setup_simple_repo(self):
        """Build the following repo

        o Isabelle (2)
        |
        | @ Added Celestine (1)
        |/
        o Initial jungle content (0)
        """
        hg = self.hgrepo
        a = self.hg_append_to

        # create a root changeset
        a('jungle', 'Babar')
        hg.commit('Initial jungle content')
        # with two children
        a('jungle', 'Celestine')
        hg.commit('Added Celestine')
        hg.update('.^')
        a('village', 'Isabelle')
        hg.commit('Added Isabelle in the village')
        # update to the first created child
        hg.update('desc(Celestine)')
        self.refresh()

    def setup_complex_repo(self):
        """build the following repo

        @ Pom is back in town
        |
        o Isabelle in the village
        |
        o Celeste in the jungle
        |
        o Babar in the jungle
        """
        hg = self.hgrepo
        a = self.hg_append_to

        def addcs(fn, content, msg):
            a(fn, content)
            hg.commit(msg)

        addcs('jungle', 'Babar', 'Babar in the jungle')
        addcs('jungle', 'Celeste', 'Celeste in the jungle')
        addcs('village', 'Isabelle', 'Isabelle in the village')
        addcs('town', 'Pom', 'Pom is back in town')

        self.cs = dict(izip(('babar', 'celeste', 'isabelle', 'pom'), count(0)))

        self.refresh()

    def amend(self, cnx, rev, date=None, msg=None):
        assert rev != '.'
        self.hg('update', str(rev))
        if msg is None:
            msg = self.all_revs(cnx)[rev].description
        args = ('-m', msg)
        if date is not None:
            args += ('-d', date)
        self.hg('commit', '--amend', *args)
        if self.hg('log', '--rev', 'unstable()'):
            self.hg('rebase', '--dest', '.', '--rev', 'unstable()')

    def accept_patch(self, cnx, rev):
        """set patch associated with rev in reviewed state"""
        revision = self.all_revs(cnx)[rev]
        patch = revision.reverse_patch_revision[0]
        iwf = patch.cw_adapt_to('IWorkflowable')
        # ensure it's in in-progress state
        self.assertEqual('in-progress', iwf.state)
        # transition to accept
        iwf.fire_transition('accept')
        patch.cw_clear_all_caches()
        iwf = patch.cw_adapt_to('IWorkflowable')
        # ensure it is now reviewed
        self.assertEqual('reviewed', iwf.state)
        cnx.commit()
        return patch

    def check_patch(self, patch, nrev, state):
        patch.cw_clear_all_caches()
        self.assertEqual(nrev, len(patch.patch_revision),
                         msg=('Wrong number of revisions for %s (%s vs %s)' %
                              (patch.dc_title(), nrev, len(patch.patch_revision))))
        iwf = patch.cw_adapt_to('IWorkflowable')
        self.assertEqual(iwf.state, state)

    def test_detect_amend(self):
        self.setup_simple_repo()
        with self.admin_access.repo_cnx() as cnx:
            # get first child (rev #1)
            patch = self.accept_patch(cnx, 1)
            # rebase the first children on top of the second, this should trigger
            # the hook and patch comparison
            self.amend(cnx, '.^', '2008-06-30', 'Added Celestine')
            cnx.commit()
            self.refresh()
            # as it is a trivial change, patches are the same, hence the state is
            # is still reviewed
            self.check_patch(patch, 2, 'reviewed')

    def test_detect_rebase(self):
        self.setup_simple_repo()
        with self.admin_access.repo_cnx() as cnx:
            # get first child (rev #1)
            patch = self.accept_patch(cnx, 1)
            # rebase the first children on top of the second, this should trigger
            # the hook and patch comparison
            self.hg('rebase')
            self.refresh()
            # as it is a trivial rebase, patches are the same, hence the state is
            # is still reviewed
            self.check_patch(patch, 2, 'reviewed')

    def test_review_life_cycle(self):
        self.setup_simple_repo()
        with self.admin_access.repo_cnx() as cnx:
            # Create a second revision, different from the first one.
            self.amend(cnx, 1, msg='Added Celestine *into the jungle*')
            self.refresh()
            patch = self.accept_patch(cnx, 3)  # 1 became 3, because of amend/rebase.
            # Rebase on top of other head, this should trigger the hook and
            # patch comparison.
            self.hg('rebase', '--dest', '2')
            self.refresh()
            # as it is a trivial rebase, patches are the same, hence the state is
            # is still reviewed
            self.check_patch(patch, 3, 'reviewed')

    def test_rebase_outdated(self):
        self.setup_simple_repo()
        with self.admin_access.repo_cnx() as cnx:
            patch = self.accept_patch(cnx, 1)
            self.check_patch(patch, 1, 'reviewed')
            patch.cw_adapt_to('IWorkflowable').fire_transition('obsolete')
            cnx.commit()
            self.check_patch(patch, 1, 'outdated')
            self.hg('rebase', '--dest', '2')
            self.refresh()
            self.check_patch(patch, 2, 'reviewed')
            wfpatch = patch.cw_adapt_to('IWorkflowable')
            self.assertEqual(wfpatch.latest_trinfo().by_transition[0].name,
                             'reopen/accept')

    def test_rebase_outdated_changed(self):
        self.setup_simple_repo()
        with self.admin_access.repo_cnx() as cnx:
            patch = self.accept_patch(cnx, 1)
            self.check_patch(patch, 1, 'reviewed')
            patch.cw_adapt_to('IWorkflowable').fire_transition('obsolete')
            cnx.commit()
            self.check_patch(patch, 1, 'outdated')
            self.hg('rebase', '--dest', '2')
            self.hg_append_to('jungle', 'Cornelius')
            self.amend(cnx, 3, msg='Added Celestine and Cornelius')
            self.refresh()
            self.check_patch(patch, 3, 'pending-review')
            wfpatch = patch.cw_adapt_to('IWorkflowable')
            self.assertEqual(wfpatch.latest_trinfo().by_transition[0].name,
                             'ask review')

    def test_detect_rebase_and_amend(self):
        self.setup_complex_repo()

        with self.admin_access.repo_cnx() as cnx:
            isabelle_patch = self.accept_patch(cnx, self.cs['isabelle'])
            babar_patch = self.accept_patch(cnx, self.cs['babar'])
            pom_patch = self.accept_patch(cnx, self.cs['pom'])
            celeste_patch = self.accept_patch(cnx, self.cs['celeste'])


            # change things
            self.amend(cnx, self.cs['isabelle'], '2000-01-01') # identical
            self.amend(cnx, self.cs['celeste'], '2001-09-11', 'plouf') # changed
            # strip intermediate ancestor of Pom
            self.hg('strip', 'desc(Pom) and obsolete() and 4:', '--hidden')

            self.refresh()

            self.check_patch(babar_patch, nrev=1, state='reviewed')
            self.check_patch(pom_patch, nrev=2, state='reviewed')
            self.check_patch(isabelle_patch, nrev=3, state='reviewed')
            self.check_patch(celeste_patch, nrev=2, state='pending-review')

    def test_amend_then_rebase(self):
        self.setup_complex_repo()

        with self.admin_access.repo_cnx() as cnx:
            isabelle_patch = self.accept_patch(cnx, self.cs['isabelle'])

            self.amend(cnx, self.cs['isabelle'], '2001-09-11', 'PLAFF') # changed
            self.amend(cnx, 'tip^', '2001-09-12', 'PLAFF') # unchanged

            self.refresh()

            self.check_patch(isabelle_patch, nrev=3, state='pending-review')


if __name__ == '__main__':
    unittest_main()



# copyright 2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-vcreview custom forms"""

__docformat__ = "restructuredtext en"

from datetime import datetime

from cubicweb.predicates import (is_instance, rql_condition, attribute_edited,
                                match_transition, match_user_groups)
from cubicweb.uilib import cut
from cubicweb.web import formfields
from cubicweb.web.views import workflow, editviews, autoform, uicfg

_affk = uicfg.autoform_field_kwargs



class AskReviewPatchChangeStateForm(workflow.ChangeStateForm):
    __select__ = is_instance('Patch') & (
        match_user_groups('managers', 'committers')
        | rql_condition('X patch_revision RE, RE from_repository R, R repository_committer U')
        ) & (
        match_transition('ask review')
        | attribute_edited('patch_reviewer'))

    patch_reviewer = autoform.etype_relation_field('Patch', 'patch_reviewer', 'subject')



class RevisionComboboxView(editviews.ComboboxView):
    __select__ = is_instance('Revision')
    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w('%s: %s' % (
            entity.dc_title(),
            cut(entity.description,
                self._cw.property_value('navigation.short-line-size'))))




import os

with dropped_constraints('Patch', 'patch_name', droprequired=True):
    add_attribute('Patch', 'patch_name')
    i = 0
    for row in rql('Patch P'):
        patch_eid = row[0]
        patch = session.entity_from_eid(patch_eid)
        directory, filename = rql('Any D,F ORDERBY R DESC LIMIT 1 '
                                  'WHERE P eid %(e)s, P patch_revision VC, VC from_revision REV, REV revision R, VC content_for VF, VF directory D, VF name F',
                                  {'e': patch_eid})[0]
        path = os.path.join(directory or '', filename)
        patch.set_attributes(patch_name=path)
        i += 1
        if i % 100 == 0:
            commit(ask_confirm=False)

commit()



# simple script that make a json dump of task and first level comment on patch
#
# run it through cubicweb-ctl shell BEFORE the update
#
# output is done on stdout. redirect it.


patch_task = '''
Any P, ST, T, D, AST
WHERE P has_activity A,
      P is Patch,
      P in_state S,
      S name ST,
      A title T,
      A description D,
      A in_state AS,
      AS name AST
'''


patch_comment = '''
Any P, ST, C
WHERE CO comments P,
      P in_state S,
      P is Patch,
      S name ST,
      CO content C
'''

vc_task = '''
Any P, ST, RE, LID, T, D, AST
WHERE P patch_revision RE,
      IP point_of RE,
      IP has_activity A,
      IP lid LID,
      P in_state S,
      P is Patch,
      S name ST,
      A title T,
      A description D,
      A in_state AS,
      AS name AST
'''

vc_comment = '''
Any P, ST, RE, LID, C
WHERE P patch_revision RE,
      IP point_of RE,
      CO comments IP,
      IP lid LID,
      P in_state S,
      P is Patch,
      S name ST,
      CO content C
'''

import re
RE_NODE= re.compile('# Node ID ([a-f0-9]{40})')

nodes_cache = {}

def get_node(patch_eid):
    if patch_eid in nodes_cache:
        return nodes_cache[patch_eid]
    vc = session.entity_from_eid(patch_eid)
    match = vc.data
    if match is not None:
        match = RE_NODE.search(match.getvalue())
        if match is not None:
            match = match.group(1)
        nodes_cache[patch_eid] = match
    return match

data = {}

for patch, state, title, comment, tstate in rql(patch_task):
    patch_key = (patch, state)
    patch_data = data.setdefault(patch_key, [])
    acti_data = {'type': 'task',
                 'title': title,
                 'comment': comment,
                 'state': tstate}
    patch_data.append((None, None, acti_data))

for patch, state, comment in rql(patch_comment):
    patch_key = (patch, state)
    patch_data = data.setdefault(patch_key, [])
    acti_data = {'type': 'comment',
                 'comment': comment,}
    patch_data.append((None, None, acti_data))

for patch, state, vc, lid, title, comment, tstate in rql(vc_task):
    patch_key = (patch, state)
    patch_data = data.setdefault(patch_key, [])
    acti_data = {'type': 'task',
                 'title': title,
                 'comment': comment,
                 'state': tstate}
    patch_data.append((get_node(vc), None, acti_data))

for patch, state, vc, lid, comment in rql(vc_comment):
    patch_key = (patch, state)
    patch_data = data.setdefault(patch_key, [])
    acti_data = {'type': 'comment',
                 'comment': comment,
                }
    patch_data.append((get_node(vc), None, acti_data))

import json
print json.dumps(data.items(), indent=4)



with dropped_constraints('InsertionPoint', 'lid', droprequired=True):
    add_attribute('InsertionPoint', 'lid')
    for ip in rql('Any X,I WHERE X is InsertionPoint, X id I').entities():
        ip.set_attributes(lid=int(ip.id))
commit()

drop_attribute('InsertionPoint', 'id')
sync_schema_props_perms('InsertionPoint')

# we have now insertion point on patch comment, needs +1 decay
rql('SET X lid LID+1 WHERE X is InsertionPoint, X lid LID')
commit()



# copyright 2011-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-vcreview specific hooks and operations"""

__docformat__ = "restructuredtext en"

import binascii
import os.path as osp
from email.Utils import parseaddr

from cubicweb import RegistryException
from cubicweb.server import hook
from cubicweb.predicates import is_instance, on_fire_transition, has_related_entities
from cubicweb.hooks import notification

from cubes.vcsfile.entities.adapters import HGRepo
from cubes.vcreview import users_by_author, compare_patches

if HGRepo.configs is None:
    HGRepo.configs = []
HGRepo.configs += ['extensions.pruned=%s/hgext/pruned.py' % osp.dirname(__file__),
                   'diff.unified=5']

def dead_revision(rev):
    # check more than hidden to protect against instability (e.g. obsolete
    # draft with alive children)
    if rev.hidden:
        return True
    if rev.phase == 'public':
        return False
    return bool(rev.reverse_obsoletes)


def rewind_state(patch):
    """Restore the state of an 'outdated' patch (if possible) to the state it
    was prior to the 'obsolete' transition.
    """
    wfable = patch.cw_adapt_to('IWorkflowable')
    trinfo = wfable.latest_trinfo()
    oldstate = trinfo.from_state[0]
    for tr in wfable.possible_transitions():
        if tr.destination_state[0] == oldstate:
            wfable.fire_transition(tr.name)
            break


class PhaseUpdate(hook.Hook):
    """detect phase/visibility changes to a changeset"""
    __regid__ = 'vcreview.phase'
    __select__ = hook.Hook.__select__ & is_instance('Revision')
    events = ('after_update_entity',)

    def __call__(self):
        repo = self.entity.repository
        if not getattr(repo, 'has_review', False): # not there during migration
            return
        rev = self.entity
        if rev.cw_edited.get('phase') == 'public':
            operation = AttachToPatchesOp.get_instance(self._cw)
            operation.add_data((repo.eid, rev.eid))
        elif (rev.cw_edited.get('hidden', False)
              and rev.reverse_patch_revision):
            patch = rev.reverse_patch_revision[0]
            operation = PatchLifeCheckOp.get_instance(self._cw)
            operation.add_data((repo.eid, patch.eid, rev.eid))


class RevisionAdded(hook.Hook):
    """detect new revision"""
    __regid__ = 'vcreview.auto-link-to-patch'
    __select__ = hook.Hook.__select__ & is_instance('Revision')
    events = ('after_add_entity',)

    def __call__(self):
        repo = self.entity.repository
        if getattr(repo, 'has_review', False): # not there during migration
            operation = AttachToPatchesOp.get_instance(self._cw)
            operation.add_data((repo.eid, self.entity.eid))


class RevisionObsoleteHook(hook.Hook):
    """detect new obsolete revision"""
    __regid__ = 'vcreview.obsolete_rev'
    __select__ = hook.Hook.__select__ & hook.match_rtype('obsoletes')
    events = ('after_add_relation',)

    def __call__(self):
        rev = self._cw.entity_from_eid(self.eidfrom)
        if rev.reverse_patch_revision:
            return
        repo = rev.repository
        if getattr(repo, 'has_review', False):
            operation = AttachToPatchesOp.get_instance(self._cw)
            operation.add_data((repo.eid, rev.eid))


class RevisionDeleted(hook.Hook):
    """When deleting a revision, delete the associated patch if it has
    no other revisions
    """
    __regid__ = 'vcreview.revision_gone'
    __select__ = (hook.Hook.__select__ & is_instance('Revision') &
                  has_related_entities('patch_revision', role='object'))
    events = ('before_delete_entity',)

    def __call__(self):
        patch = self.entity.reverse_patch_revision[0]
        DeletedRevisionCheckOp.get_instance(self._cw).add_data(patch.eid)


class AttachToPatchesOp(hook.DataOperationMixIn, hook.LateOperation):
    """Attach new revision to a Patch entity

    Create new Patches as necessary.
    """
    def precommit_event(self):
        for eid_repo, eid_rev in sorted(self.get_data()):
            # sorted for topological ordering
            # Helps to have more correct result (but no guarantee)
            repo = self.cnx.entity_from_eid(eid_repo)
            rev = self.cnx.entity_from_eid(eid_rev)
            operation = PatchLifeCheckOp.get_instance(self.cnx)
            patch = None
            if rev.reverse_patch_revision:
                patch = rev.reverse_patch_revision[0]
            else:
                patches = self.related_patch(rev)
                if patches:
                    patch = patches[0]
                    rev.cw_set(reverse_patch_revision=patch.eid)
                    for folded in patches[1:]:
                        operation.add_data((repo.eid, folded.eid, rev.eid))
                elif rev.phase != 'public':
                    patch = self.create_patch(rev)
            if patch is not None:
                # none if new public revison spawn unrelated to any patch
                users = users_by_author(self.cnx.execute, rev.author)
                if users:
                    self.cnx.execute('SET P owned_by U WHERE P eid %%(p)s, U eid IN (%s), NOT P owned_by U' %
                                     ','.join([str(u) for u in users]),
                                     {'p': patch.eid})
                operation.add_data((repo.eid, patch.eid, rev.eid))

    def related_patch(self, rev):
        patches = []
        for prec in sorted(rev.obsoletes, key=lambda e: e.eid):
            if prec.reverse_patch_revision:
                patch = prec.reverse_patch_revision[0]
                if patch.is_final:
                    continue
                # is every revision dead?
                tip = patch.tip()
                dead_tip = True
                if tip is None or dead_revision(tip):
                    patches.append(patch)
                else:
                    msg = '%s: %s not attached to Patch %i. %s is still alive'
                    self.info(msg, rev.repository.dc_title(),
                                   rev.changeset,
                                   patch.eid,
                                   tip.changeset)

        return patches


    def create_patch(self, rev):
        name = (rev.description.splitlines() or [rev.dc_title()])[0]
        kwargs = {'patch_name': name, 'patch_revision': rev}
        users = users_by_author(self.cnx.execute, rev.author)
        if users:
            kwargs['created_by'] = users[0]
        patch = self.cnx.create_entity('Patch', **kwargs)
        msg = '%s: New patch %i for new draft revision %s'
        self.info(msg, rev.repository.dc_title(), patch.eid, rev.changeset)
        return patch


class DeletedRevisionCheckOp(hook.DataOperationMixIn, hook.Operation):
    """Delete patches with no remaining revision"""
    def precommit_event(self):
        for patch_eid in self.get_data():
            rset = self.cnx.execute(
                    'DELETE Patch P WHERE P eid %(eid)s,'
                    '                     NOT EXISTS(P patch_revision R)',
                    {'eid': patch_eid})
            if rset:
                self.info('Deleted patch %s, no revisions left', rset[0][0])


class PatchLifeCheckOp(hook.DataOperationMixIn, hook.SingleLastOperation):
    """Put patches in the right state

    invoke it with:

        (repository-eid, patch-eid, revision-eid)

    The revision has "some new activity" and is related to the patch.  There is
    usually a PATCH patch_revision REVISION relation between them except in the
    folded case.
    """

    FOLDED = ('Any OLD '
              'WHERE P eid %(p)s, '
              '      R eid %(r)s, '
              '      NOT P patch_revision R, '
              '      P patch_revision OLD, '
              '      R obsoletes OLD')

    def precommit_event(self):
        cnx = self.cnx
        for eid_repo, eid_patch, eid_rev in sorted(self.get_data()):
            patch = cnx.entity_from_eid(eid_patch)
            wfable = patch.cw_adapt_to('IWorkflowable')
            ### Case 1: patch is already in final state
            # nothing to be done on it anymore
            if patch.is_final:
                continue
            rev = cnx.entity_from_eid(eid_rev)
            assert rev.reverse_patch_revision
            ### Case 2: patch has just been applied
            if (rev.phase == 'public'
                and rev.reverse_patch_revision[0].eid == patch.eid):
                # there is one applied version related to this patch
                wfable.fire_transition('apply', u'applied as %s'
                                       % rev.changeset)
                continue
            # get the patch's tips
            data = {'p': eid_patch, 'r': rev.eid}
            tips_rset = patch.all_tips()
            visible_tips = set(r[0] for r in tips_rset if not r[1])
            if visible_tips:
                # pre-existing revisions are likely just set to hidden=True
                # in this transaction. We do not want them to affect patch
                # state if we have a new visible tip (or even newer revision).
                if not (cnx.added_in_transaction(eid_rev)
                        or eid_rev in visible_tips):
                    continue
                same_patch, ancestor = self.compare_with_ancestor(
                    rev, wfable.state)
                if same_patch:
                    if wfable.state == 'outdated':
                        rewind_state(patch)
                    continue
                ### Case 3: new version of the patch
                name = (rev.description.splitlines() or [rev.dc_title()])[0]
                if name != patch.patch_name:
                    patch.cw_set(patch_name=name)
                name, email = parseaddr(rev.author)
                auto = cnx.execute('Any V WHERE U primary_email E, '
                                   'E address ILIKE %(email)s, '
                                   'P for_user U, '
                                   'P pkey "vcreview.autoreview", '
                                   'P value V',
                                   {'email': email.decode('ascii')})
                # defaults to auto ask review
                autoreview = auto and auto[0][0] == "1"
                if autoreview:
                    if wfable.state != 'pending-review':
                        wfable.fire_transition('ask review', u'new version %s'
                                               % rev.changeset)
                else:
                    if wfable.state == 'outdated':
                        wfable.fire_transition('reopen', u'new version %s'
                                               % rev.changeset)
                    elif wfable.state == 'reviewed':
                        msg = u'new version %s' % rev.changeset
                        if ancestor is not None:
                            msg += u' (different from :eid:`%d:%s`)' % (
                                    ancestor.eid, ancestor.changeset)
                        wfable.fire_transition('ask review', msg, u'text/rest')
            # patch has no visible tip
            elif cnx.execute(self.FOLDED, data):
                ### Case 4: Patch is now folded somewhere else
                wfable.fire_transition('fold', u'folded into %s'
                                       % rev.changeset)
            else:
                # all tips are hidden otherwise we would have walked in the
                # visible_tips conditional
                if eid_rev not in set(r[0] for r in tips_rset):
                    # we do not care about this revision
                    continue
                if rev.reverse_obsoletes:
                    # the patch will need a fold later
                    continue
                repo = cnx.entity_from_eid(eid_repo)
                succ = None
                for tip_eid, _ in tips_rset:
                    tip = cnx.entity_from_eid(tip_eid)
                    succ = self.has_successors(repo, tip)
                    if succ:
                        break
                if not succ:
                    ### Case 5: All tips are plain pruned. reject
                    wfable.fire_transition('reject', u'all versions pruned')
                elif wfable.state != 'outdated':
                    ### Case 6: Patch has a new version somewhere else
                    msg = u'outdated'
                    wfable.fire_transition('obsolete', msg)

    def has_successors(self, repo, rev):
        """return true if there is a final successor recorded in evolution history"""
        with repo.cw_adapt_to('VCSRepo').hgrepo() as localrepo:
            return bool(localrepo.log(revrange='%s and not pruned()' % rev.changeset))

    def compare_with_ancestor(self, rev, patch_state):
        """Compare specified revision with its most recent ancestor and return
        (True, None) if they are similar else (False, ancestor). If the
        revision has no ancestor, (False, None) is returned.
        """
        # XXX Why do we need this guard?
        if patch_state in ('in-progress', 'pending-review', 'reviewed', 'outdated'):
            rset = self.cnx.execute('Any R ORDERBY MD DESC LIMIT 1 WHERE'
                                    ' X obsoletes R, R modification_date MD,'
                                    ' X eid %(x)s', {'x': rev.eid})
            if rset:
                ancestor = rset.get_entity(0, 0)
                if not compare_patches(ancestor.export(), rev.export()):
                    return False, ancestor
                return True, None
        return False, None


class InitPatchCreatorHook(hook.Hook):
    """automatically add people to nosy list

    Revision added to a patch should add the author to nosy list if it's known.

    This is done by searching author's email, then searching for a user
    associated to this email.
    """
    __regid__ = 'vcreview.auto-nosy'
    __select__ = hook.Hook.__select__ & hook.match_rtype('patch_revision')
    events = ('after_add_relation',)
    category = 'metadata'

    def __call__(self):
        patch = self._cw.entity_from_eid(self.eidfrom)
        author = self._cw.entity_from_eid(self.eidto).author
        ueids = users_by_author(self._cw.execute, author)
        for ueid in ueids:
            self._cw.execute(
                'SET P nosy_list U WHERE P eid %(p)s, U eid %(u)s,'
                'NOT P nosy_list U', {'p': patch.eid, 'u': ueid})
        # legacy code about created by. dropped for now.
        # self._cw.add_relations([('created_by', relations),
        #                         ('owned_by', relations)])




class SetReviewerHook(hook.Hook):
    """register an operation to set a reviewer when the 'ask review' transition
    is fired
    """
    __regid__ = 'vcreview.set-reviewer'
    __select__ = hook.Hook.__select__ & on_fire_transition('Patch', 'ask review')
    events = ('after_add_entity',)

    def __call__(self):
        SetReviewerOp.get_instance(self._cw).add_data(self.entity)


class SetReviewerOp(hook.DataOperationMixIn, hook.Operation):

    def precommit_event(self):
        for trinfo in self.get_data():
            patch = trinfo.for_entity
            patch.cw_adapt_to('IPatchReviewControl').set_reviewers()


class ReviewerSetHook(hook.Hook):
    """send a notification when a user is assigned to review a patch and ensure
    he is in the patch nosy-list
    """
    __regid__ = 'vcreview.reviewer-set'
    __select__ = hook.Hook.__select__ & hook.match_rtype('patch_reviewer')
    events = ('after_add_relation',)

    def __call__(self):
        # send notification
        rset = self._cw.eid_rset(self.eidfrom)
        try:
            view = self._cw.vreg['views'].select(
                'vcreview.notifications.reviewer-set', self._cw,
                rset=rset, user=self._cw.eid_rset(self.eidto).get_entity(0, 0))
        except RegistryException:
            # view unregistered or unselectable
            return
        notification.notify_on_commit(self._cw, view=view)
        # ensure user is in the nosy-list
        self._cw.execute('SET P nosy_list U '
                         'WHERE P eid %(p)s, U eid %(u)s, NOT P nosy_list U',
                         {'p': self.eidfrom, 'u': self.eidto})


class AddTaskAuthorToPatchNosylistHook(hook.Hook):
    """Add the author of a task to the patch's nosylist"""
    __regid__ = 'vcreview.add-task-author-to-patch-nosylist'
    __select__ = (hook.Hook.__select__ &
                  hook.match_rtype('has_activity'))
    events = ('after_add_relation',)

    def __call__(self):
        patch = self._cw.entity_from_eid(self.eidto).patch
        AddUserToPatchNosyListOp.get_instance(self._cw).add_data(
            (patch.eid, self.eidto))


class AddTaskCommenterToPatchNosylistHook(hook.Hook):
    """Add users commenting a task to the patch's nosylist"""
    __regid__ = 'vcreview.add-task-commenter-to-patch-nosylist'
    __select__ = (hook.Hook.__select__ &
                  hook.match_rtype('comments', toetypes=('Task', 'Comment')))
    events = ('after_add_relation',)

    def __call__(self):
        comment = self._cw.entity_from_eid(self.eidfrom)
        comment_root = comment.cw_adapt_to('ITree').root()
        if comment_root.cw_etype != 'Task':
            return
        patch = comment_root.patch
        if not patch:
            return
        # `created_by` relation may not be set, so proceed in an operation
        AddUserToPatchNosyListOp.get_instance(self._cw).add_data(
            (patch.eid, comment.eid))


class AddUserToPatchNosyListOp(hook.DataOperationMixIn, hook.Operation):
    """Add the creator of a task/comment to the patch nosylist"""

    def precommit_event(self):
        for patch_eid, task_or_comment_eid in self.get_data():
            self.cnx.execute('SET P nosy_list U '
                             'WHERE P eid %(p)s, NOT P nosy_list U,'
                             '      TC created_by U, TC eid %(tc)s',
                             {'p': patch_eid, 'tc': task_or_comment_eid})


from cubes.nosylist import hooks as nosylist
nosylist.S_RELS |= set(('has_activity', 'patch_revision'))
nosylist.O_RELS |= set(('point_of', 'comments'))



wf = get_workflow_for('Task')
for name in ('done', 'start'):
    tr = wf.transition_by_name(name)
    for cond in tr.condition:
        expr = cond.expression.replace('P patch_repository R', 'P patch_revision RE, RE from_repository R')
        expr = expr.replace('IP point_of VC, P patch_revision VC', 'IP point_of RE, P patch_revision RE')
        if expr != cond.expression:
            cond.set_attributes(expression=expr)

commit()



sync_schema_props_perms('Patch', syncprops=False)
sync_schema_props_perms('patch_repository', syncprops=False)
sync_schema_props_perms('patch_revision', syncprops=False)
sync_schema_props_perms('applied_at', syncprops=False)


pwf = rql('Workflow X WHERE X name "patch workflow"').get_entity(0, 0)
folded  = pwf.add_state(_('folded'))
inprogress = pwf.state_by_name('in-progress')
pending = pwf.state_by_name('pending-review')
deleted = pwf.state_by_name('deleted')
pwf.add_transition(_('fold'), (inprogress, pending, deleted), folded,
                   ('managers',),)

